"""
Wrapper for ``bdinfo`` command
"""

import functools

from .. import errors, utils
from ..utils import LazyModule, daemon
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = LazyModule(module='natsort', namespace=globals())


class PlaylistsJob(base.JobBase):
    """
    Ask the user to select playlists from Blu-ray or DVD disc(s)

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``discs_available``
            Emitted when all subdirectories containing "BDMV" directories are found. Registered
            callbacks get a sequence of directory paths that each contain a "BDMV" directory.

        ``discs_selected``
            Emitted when the user has selected Blu-ray path(s). Registered callbacks get a sequence
            of directory paths that each contain a "BDMV" directory.

        ``playlists_available``
            Emitted when the list of playlists is available. Registered callbacks get a dictionary
            with they keys ``disc_path``, the parent path of the "BDMV" directory, and
            ``playlists``, a sequence of playlist dictionaries. Each playlist dictionary has the
            keys ``filepath`` ``size``, ``duration``, ``disc_path`` and ``items``. ``items`` is a
            sequence of ``.m2ts`` file paths. Playlists and playlist items are sorted by ``size`` in
            reverse order.

        ``playlists_selected``
            Emitted when the user has selected playlist(s). Registered callbacks get a sequence of
            playlist ``items`` as described in ``playlists_available``.
    """

    name = 'playlists'
    label = 'Playlists'

    @functools.cached_property
    def cache_id(self):
        """Final segment of `content_path` and "select_multiple" argument"""
        cache_id = [
            utils.fs.basename(self._content_path),
        ]
        if self.select_multiple:
            cache_id.append('select_multiple')
        return cache_id

    @property
    def select_multiple(self):
        """Whether the user may select multiple playlists"""
        return self._select_multiple

    def initialize(self, *, content_path, select_multiple=False):
        """
        Set internal state

        :param content_path: Path to directory that contains "BDMV" or "VIDEO_TS" directory

            May also contain multiple subdirectories that contain "BDMV" or "VIDEO_TS" directories
            (multidisc)

        :param bool select_multiple: Whether the user may select multiple playlists
        """
        self._content_path = content_path
        self._select_multiple = bool(select_multiple)
        self._selected_discpaths = set()
        self._processed_discpaths = set()
        self._selected_playlists = []
        self._playlists_process = None

        self.signal.add('discs_available')
        self.signal.add('discs_selected')
        self.signal.add('playlists_available')
        self.signal.add('playlists_selected', record=True)
        self.signal.register('playlists_selected', self._store_selected_playlists)

    async def run(self):
        self._playlists_process = daemon.DaemonProcess(
            name=self.name,
            target=_playlists_process,
            kwargs={
                'content_path': self._content_path,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._playlists_process.start()
        await self._playlists_process.join()
        await self.finalization()

        for playlist in self._selected_playlists:
            self.add_output(playlist.filepath)

    def terminate(self, reason=None):
        if self._playlists_process:
            self._playlists_process.stop()
        super().terminate(reason=reason)

    @base.unless_job_is_finished
    def _handle_info(self, info):
        if 'discs_available' in info:
            # Don't ask the user to pick discs if there is only one.
            discpaths = info['discs_available']
            for discpath in discpaths:
                _log.debug('Disc found: %r', discpath)
            if len(discpaths) <= 1:
                self.discs_selected(discpaths)
            else:
                self.signal.emit('discs_available', discpaths)

        elif 'playlists_available' in info:
            discpath, playlists = info['playlists_available']
            # Don't ask the user to pick discs if there is only one.
            for playlist in playlists:
                _log.debug('Playlist found: %r', playlist)
            if len(playlists) <= 1:
                self.playlists_selected(discpath, playlists)
            else:
                self.signal.emit('playlists_available', discpath, playlists)

        else:
            raise RuntimeError(f'Unexpected info: {info!r}')

    @base.unless_job_is_finished
    def _handle_error(self, error):
        if isinstance(error, errors.DependencyError):
            self.error(error)
        else:
            self.exception(error)

    def discs_selected(self, discpaths):
        """
        Called by the UI when the user selects disc(s)

        :param str discpaths: Sequence of directory paths that contain a "BDMV" or "VIDEO_TS"
            directory

        Calling this method emits the ``discs_selected`` signal.
        """
        for discpath in discpaths:
            _log.debug('Disc selected: %r', discpath)
        self._selected_discpaths.update(discpaths)
        self._playlists_process.send(daemon.MsgType.info, {'discs_selected': discpaths})
        self.signal.emit('discs_selected', discpaths)

        # If the user deselected all discs, playlists_selected() will never be called, which is
        # supposed to call finalize() to finish the job, so we must call finalize() now.
        if not discpaths:
            self.finalize()

    def playlists_selected(self, discpath, playlists):
        """
        Called by the UI when the user selects playlist(s)

        :param str discpath: Directory path that contains the `playlists`
        :param playlists: Sequence of :class:`~.types.Playlist` instances

        Calling this method emits the ``playlists_selected`` signal.
        """
        for playlist in playlists:
            _log.debug('Playlist selected: %r', playlist)

        self.signal.emit('playlists_selected', discpath, playlists)
        self._processed_discpaths.add(discpath)

        # Check if playlist(s) were selected for every selected disc. Note that it is possible
        # to select zero playlists from any disc or all discs.
        if self._processed_discpaths == self._selected_discpaths:
            _log.debug('All disc paths processed: %r', self._processed_discpaths)
            self.finalize()

        # If we are only interested in one disc and the user selected at least one playlist from
        # this disc, we don't need to prompt for more disks.
        elif playlists and not self.select_multiple:
            _log.debug('No more playlists required - not prompting user for any remaining discs')
            self.finalize()

    def _store_selected_playlists(self, discpath, playlists):
        self._selected_playlists.extend(playlists)

    @property
    def selected_playlists(self):
        """
        Sequence of :class:`~.types.Playlist` instances that were selected by the user

        This sequence will be empty at the beginning and grow as the user makes selections for each
        selected disc.
        """
        return tuple(self._selected_playlists)


def _playlists_process(output_queue, input_queue, *, content_path):
    if utils.disc.is_bluray(content_path, multidisc=True):
        disc_module = utils.disc.bluray
    elif utils.disc.is_dvd(content_path, multidisc=True):
        disc_module = utils.disc.dvd
    else:
        raise errors.ContentError(f'No BDMV or VIDEO_TS subdirectory found: {content_path}')

    selected_discs = _get_selected_discs(
        output_queue, input_queue,
        content_path=content_path,
        disc_module=disc_module,
    )
    for discpath in selected_discs:
        utils.daemon.maybe_terminate(input_queue)
        _report_available_playlists(output_queue, discpath=discpath, disc_module=disc_module)


def _get_selected_discs(output_queue, input_queue, *, content_path, disc_module):
    # Find all discs (in case of multi-disc releases).
    discs_available = tuple(natsort.natsorted(disc_module.get_disc_paths(content_path)))
    if not discs_available:
        output_queue.put((utils.daemon.MsgType.error, errors.ContentError(f'No disc found: {content_path}')))
        return ()
    else:
        # Report available discpaths back to main process.
        output_queue.put((utils.daemon.MsgType.info, {'discs_available': discs_available}))
        # Get user-selected disc(s) from main process.
        return utils.daemon.read_input_queue_key(input_queue, 'discs_selected')


def _report_available_playlists(output_queue, *, discpath, disc_module):
    playlists_available = _extend_playlists_info(disc_module.get_playlists(discpath))
    output_queue.put((utils.daemon.MsgType.info, {'playlists_available': (discpath, playlists_available)}))


def _extend_playlists_info(playlists):
    # Sort playlists in natural sort order by file path.
    return tuple(natsort.natsorted(playlists, key=lambda playlist: playlist.filepath))



"""
Concrete :class:`~.TrackerBase` subclass for PTP
"""

import asyncio
import math
import re
import urllib

from ... import errors, utils
from ..base import TrackerBase
from . import rules
from .config import PtpTrackerConfig
from .jobs import PtpTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTracker(TrackerBase):
    name = 'ptp'
    label = 'PTP'
    torrent_source_field = 'PTP'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. Login Credentials\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.username USERNAME\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.password PASSWORD\n'
        '\n'
        '{howto.next_section}. Announce URL\n'
        '\n'
        '   The announce URL is required because the passkey is used for 2FA.\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.announce_url ANNOUNCE_URL\n'
        '\n'
        '{howto.next_section}. Screenshots\n'
        '\n'
        '   {howto.current_section}.1 Configure ptpimg.me API\n'
        '       $ upsies upload-images ptpimg --help\n'
        '\n'
        '   {howto.current_section}.2 Specify how many screenshots to make. (optional)\n'
        '       $ upsies set trackers.{tracker.name}.screenshots_from_movie NUMBER_OF_MOVIE_SCREENSHOTS\n'
        '       $ upsies set trackers.{tracker.name}.screenshots_from_episode NUMBER_OF_EPISODE_SCREENSHOTS\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = PtpTrackerConfig
    TrackerJobs = PtpTrackerJobs
    rules = rules

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _ajax_url(self):
        return urllib.parse.urljoin(self._base_url, '/ajax.php')

    @property
    def _artist_url(self):
        return urllib.parse.urljoin(self._base_url, '/artist.php')

    @property
    def _logout_url(self):
        return urllib.parse.urljoin(self._base_url, '/logout.php')

    @property
    def _upload_url(self):
        return urllib.parse.urljoin(self._base_url, '/upload.php')

    @property
    def _torrents_url(self):
        return urllib.parse.urljoin(self._base_url, '/torrents.php')

    @property
    def _announce_url(self):
        announce_url = self.options['announce_url']
        if not announce_url:
            raise errors.AnnounceUrlNotSetError(tracker=self)
        else:
            return self.options['announce_url']

    @property
    def _passkey(self):
        # Needed for logging in with ajax.php
        match = re.search(r'.*/([a-zA-Z0-9]+)/announce', self._announce_url)
        if match:
            return match.group(1)
        else:
            raise RuntimeError(f'Failed to find passkey in announce URL: {self._announce_url}')

    async def _request(self, method, *args, error_prefix='', **kwargs):
        # Because HTTP errors (e.g. 404) are raised, we treat RequestErrrors as
        # normal response so we can get the message from the HTML.
        try:
            # `method` is "GET" or "POST"
            response = await getattr(utils.http, method.lower())(
                *args,
                user_agent=True,
                follow_redirects=False,
                cookies=self.cookies_filepath,
                **kwargs,
            )
        except errors.RequestError as e:
            response = e

        # Get error from regular exception (e.g. "Connection refused") or the
        # HTML in response.
        try:
            self._maybe_raise_error(response)
        except errors.RequestError as e:
            # Prepend error_prefix to explain the general nature of the error.
            if error_prefix:
                raise errors.RequestError(f'{error_prefix}: {e}') from e
            else:
                raise e
        else:
            return response

    def _maybe_raise_error(self, response_or_request_error):
        # utils.http.get()/post() raise RequestError on HTTP status codes, but
        # we want to get the error message from the response text.
        # _maybe_raise_error_from_*() handle Response and RequestError.
        self._maybe_raise_error_from_json(response_or_request_error)
        self._maybe_raise_error_from_html(response_or_request_error)

        # If we got a RequestError and we didn't find an error message in the
        # text, we raise it. This handles any real RequestErrors, like
        # "Connection refused". We also raise any other exception so _request()
        # doesn't return it as a regular response.
        if isinstance(response_or_request_error, BaseException):
            raise response_or_request_error

    def _maybe_raise_error_from_json(self, response_or_request_error):
        # Get error message from ajax.php JSON Response or RequestError
        try:
            json = response_or_request_error.json()
        except errors.RequestError:
            # Response or RequestError is not JSON
            pass
        else:
            if (
                isinstance(json, dict)
                and json.get('Result') == 'Error'
                and json.get('Message')
            ):
                raise errors.RequestError(utils.html.as_text(json['Message']))

    def _maybe_raise_error_from_html(self, response_or_request_error):
        # Only attempt to find an error message if this looks like HTML. This
        # prevents a warning from bs4 about parsing non-HTML.
        text = str(response_or_request_error)
        if all(c in text for c in '<>\n'):
            doc = utils.html.parse(text)
            try:
                error_header_tag = doc.select('#content .page__title', string=re.compile(r'(?i:error)'))
                error_container_tag = error_header_tag[0].parent
                error_msg_tag = error_container_tag.find('div', attrs={'class': 'panel__body'})
                error_msg = error_msg_tag.get_text().strip()
                if error_msg:
                    raise errors.RequestError(error_msg)
            except (AttributeError, IndexError):
                # No error message found
                pass

    async def _get_auth(self):
        response = await self._request('GET', self._base_url)
        doc = utils.html.parse(response)
        auth_regex = re.compile(r'logout\.php\?.*\bauth=([0-9a-fA-F]+)')
        logout_link_tag = doc.find('a', href=auth_regex)
        if logout_link_tag:
            logout_link_href = logout_link_tag['href']
            match = auth_regex.search(logout_link_href)
            return match.group(1)

        raise errors.RequestError('Could not find auth')

    async def _login(self, *, tfa_otp=None):
        if not self.options.get('username'):
            raise errors.RequestError('Login failed: No username configured')
        elif not self.options.get('password'):
            raise errors.RequestError('Login failed: No password configured')

        post_data = {
            'username': self.options['username'],
            'password': self.options['password'],
            'passkey': self._passkey,
            # Do not send Tfa* on the first request. (`None` are exluded from the POST data.)
            'TfaType': 'normal' if tfa_otp else None,
            'TfaCode': tfa_otp,
            # Tell the server to remember our user session if the user wants to store it.
            'keeplogged': '1' if self.cookies_filepath else None,
        }

        response = await self._request(
            method='POST',
            url=self._ajax_url,
            params={'action': 'login'},
            data=post_data,
            error_prefix='Login failed',
        )
        json = response.json()

        if json.get('Result') == 'TfaRequired':
            # Raise exception to get called again with `tfa_otp`.
            raise errors.TfaRequired(f'2FA OTP required: {response.json()!r}')

        elif json.get('Result') == 'Error':
            if msg := json.get('Message'):
                raise errors.RequestError(f'Login failed: {utils.html.as_text(msg)}')
            else:
                raise errors.RequestError('Login failed')

        else:
            self._anti_csrf_token = response.json()['AntiCsrfToken']

    async def confirm_logged_in(self):
        await self._get_auth()

    async def _logout(self):
        try:
            response = await self._request(
                method='GET',
                url=self._logout_url,
                params={'auth': await self._get_auth()},
                error_prefix='Logout failed',
            )
            _log.debug('Logout response: %r', response)
        finally:
            if hasattr(self, '_anti_csrf_token'):
                delattr(self, '_anti_csrf_token')

    async def get_announce_url(self):
        return self._announce_url

    async def upload(self, tracker_jobs):
        post_data = tracker_jobs.post_data
        post_data['AntiCsrfToken'] = self._anti_csrf_token

        _log.debug('POSTing data:')
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        post_files = {
            'file_input': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }
        _log.debug('POSTing files: %r', post_files)

        response = await utils.http.post(
            url=self._upload_url,
            cache=False,
            user_agent=True,
            data=post_data,
            files=post_files,
            # Ignore the HTTP redirect (should be 302 Found) so we can get the
            # torrent URL from the "Location" response header
            follow_redirects=False,
        )

        # The server needs some time to process the uploaded metadata before the
        # torrent is registered by the tracker and we can seed it. If we don't
        # sleep() here, the torrent may be announced before the tracker knows it
        # exists.
        await asyncio.sleep(1)

        return self._handle_upload_response(response)

    def _handle_upload_response(self, response):
        # "Location" header should contain the uploaded torrent's URL
        _log.debug('Upload response headers: %r', response.headers)
        location = response.headers.get('Location')
        _log.debug('Upload response location: %r', location)
        if location:
            torrent_page_url = urllib.parse.urljoin(self.options['base_url'], location)
            # Redirect URL should start with "https://.../torrents.php"
            if torrent_page_url.startswith(self._torrents_url):
                return torrent_page_url

        # Find error message in HTML
        doc = utils.html.parse(response)

        # Find and raise error message
        alert_tag = doc.find(class_='alert')
        if alert_tag:
            msg = utils.html.as_text(alert_tag)
            raise errors.RequestError(f'Upload failed: {msg}')

        # Failed to find error message
        dump_filepath = 'ptp_upload_failed.html'
        utils.html.dump(response, dump_filepath)
        raise errors.RequestError(f'Failed to interpret response (see {dump_filepath})')

    def normalize_imdb_id(self, imdb_id):
        """
        Format IMDb ID for PTP

        PTP expects 7-characters, right-padded with "0" and without the leading
        "tt".

        If `imdb_id` is falsy (e.g. `None`, empty string, etc), or if it isn't
        an IMDb ID, return "0".
        """
        imdb_id = str(imdb_id)
        match = re.search(r'^(?:tt|)(\d+)$', imdb_id)
        if match:
            imdb_id_digits = match.group(1)
            if set(imdb_id_digits) != {'0'}:
                return match.group(1).rjust(7, '0')
        return '0'

    async def get_ptp_group_id_by_imdb_id(self, imdb_id):
        """
        Convert IMDb ID to PTP group ID

        Any :class:`~.RequestError` is caught and passed to
        :meth:`.TrackerBase.error`.

        :return: PTP group ID or `None` if PTP doesn't have a group for
            `imdb_id`
        :raise RequestError: if the request fails.
        """
        if imdb_id:
            _log.debug('%s: Fetching PTP group ID', imdb_id)
            # We must be logged in first.
            await self.signal.wait_for('logged_in')

            response = await self._request(
                method='GET',
                url=self._torrents_url,
                params={
                    'imdb': self.normalize_imdb_id(imdb_id),
                    'json': '1',
                },
                cache=True,
            )

            match = re.search(r'id=(\d+)', response.headers.get('location', ''))
            if match:
                _log.debug('%s: PTP group ID: %s', imdb_id, match.group(1))
                return match.group(1)
            else:
                _log.debug('%s: No PTP group ID', imdb_id)

    async def get_movie_metadata(self, imdb_id):
        """
        Get metadata about movie as :class:`dict` from PTP website

        :param imdb_id: IMDb ID (e.g. ``tt123456``)

        The returned :class:`dict` always has the following keys:

            - ``countries`` (:class:`tuple` of :class:`str`)
            - ``languages`` (:class:`tuple` of :class:`str`)
            - ``plot`` (:class:`str`)
            - ``poster`` (:class:`str`, image URL)
            - ``tags`` (:class:`tuple` of :class:`str`)
            - ``title`` (:class:`str`)
            - ``year`` (:class:`str`)

        :raise RequestedNotFoundError: if no metadata is found.
        :raise RequestError: if the request fails.
        """
        _log.debug('%s: Fetching metadata from PTP', imdb_id)
        if imdb_id:
            # We must be logged in first.
            await self.signal.wait_for('logged_in')
            response = await self._request(
                method='GET',
                url=self._ajax_url,
                params={
                    'action': 'torrent_info',
                    'imdb': self.normalize_imdb_id(imdb_id),
                },
                cache=True,
            )

            # We expect a list of JSON objects that contains exactly one JSON
            # object.
            results = response.json()
            _log.debug('%s: Raw movie metadata: %r', imdb_id, response)
            assert len(results) == 1, results
            metadata = results[0]

            # If `imdb_id` is invalid/unknown, "title" is `null` in the JSON
            # response, which is translated to `None`.
            if not metadata.get('title'):
                raise errors.RequestedNotFoundError(imdb_id)

        else:
            metadata = {}

        def comma_split(string):
            lst = []
            for val in string.split(','):
                val = val.strip()
                if val:
                    lst.append(val)
            return lst

        return {
            'title': metadata.get('title') or '',
            'plot': metadata.get('plot') or '',
            'poster': metadata.get('art') or '',
            'year': metadata.get('year') or '',
            'tags': comma_split(metadata.get('tags', '')),
            'countries': comma_split(metadata.get('Countries', '')),
            'languages': comma_split(metadata.get('Languages', '')),
        }

    async def get_artist_metadata(self, artist):
        """
        Get metadata about artist as :class:`dict` from PTP website

        :param artist: Name, IMDb URL/ID (e.g. ``nm123456``) or PTP URL/ID

        :raise RequestedNotFoundError: if `artist` is unknown on both PTP and
            IMDb.
        :raise RequestError: if the metadata request fails.
        """
        _log.debug('%s: Fetching artist metadata from PTP', artist)
        # We must be logged in first.
        await self.signal.wait_for('logged_in')
        try:
            response = await self._request(
                method='POST',
                url=self._artist_url,
                data={
                    'action': 'find',
                    'name': (
                        # PTP accepts PTP artist URLs, but not a naked ID.
                        self._get_artist_url(artist)
                        if re.search(r'^\d+$', artist) else
                        artist
                    ),
                    'AntiCsrfToken': self._anti_csrf_token,
                },
                cache=True,
            )
        except errors.RequestError as e:
            if 'not found' in str(e):
                raise errors.RequestedNotFoundError(artist) from e
            else:
                raise e
        else:
            return self._get_artist_dict(response)

    async def create_artist(self, name):
        """
        Create artist on PTP

        .. warning:: This must only be called if the artist doesn't exist on
            IMDb or PTP.

        :param name: Complete canonical name of the artist

        :return: :class:`dict` with the keys: ``name``, ``id``, ``url``

        :raise RequestError: if the metadata request fails.
        """
        _log.debug('Creating artist: %r', name)
        # We must be logged in first.
        await self.signal.wait_for('logged_in')
        response = await self._request(
            method='POST',
            url=self._artist_url,
            data={
              'action': 'create',
              'name': name,
              'AntiCsrfToken': self._anti_csrf_token,
            },
            cache=False,
        )
        return self._get_artist_dict(response)

    def _get_artist_dict(self, response):
        _log.debug('Raw artist metadata: %r', response)
        # Raise RequestError if response is not valid JSON.
        artist = response.json()
        try:
            return {
                'name': artist['ArtistName'],
                'id': artist['ArtistId'],
                'url': self._get_artist_url(artist['ArtistId'])
            }
        except KeyError as e:
            if artist.get('Message'):
                raise errors.RequestError(artist['Message']) from e
            else:
                raise errors.RequestError(f'Unexpected response: {artist}') from e

    def _get_artist_url(self, artist_id):
        return self._artist_url + '?id=' + str(artist_id)

    @staticmethod
    def calculate_piece_size(bytes):
        """
        Return the recommended piece size for a given content size

        :param bytes: Torrent's content size without any excluded files
        """
        exponent = math.ceil(math.log2(bytes / 1050))
        # Allowed piece size range: 32 KiB ... 16 MiB
        exponent = max(15, min(24, exponent))
        return int(math.pow(2, exponent))

    @staticmethod
    def calculate_piece_size_min_max(bytes):
        """
        Return the allowed minimum and maximum piece size for a given
        content size

        :param bytes: torrent's content size without any excluded files

        :raise ValueError: if `bytes` is negative or otherwise unexpected
        """
        if bytes <= 0:
            raise ValueError(f'Unexpected size: {bytes!r}')

        # NOTE: The algorithm below is from the website's javascript.

        max_exponent = math.ceil(math.log2(bytes / 500))
        min_exponent = math.floor(math.log2(bytes / 2000))

        # 2^24 or 16MiB is the max for uTorrent 2.x
        min_exponent = min(24, min_exponent)
        max_exponent = min(24, max_exponent)

        # Compatibility for uTorrent 2.x creator which only supports up to 4 MiB piece sizes
        if min_exponent > 22 and bytes <= 128 * 1024 * 1024 * 1024:
            min_exponent = 22
        if min_exponent > 22 and bytes <= 256 * 1024 * 1024 * 1024:
            min_exponent = 23

        # Tiny torrents < ~40MiB need a sane lower bound
        min_exponent = max(15, min_exponent)
        max_exponent = max(18, max_exponent)

        return (
            math.pow(2, min_exponent),
            math.pow(2, max_exponent),
        )



import functools
import os

from .. import widgets
from . import JobWidgetBase


class MediainfoJobWidget(JobWidgetBase):

    is_interactive = False

    def setup(self):
        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
            extend_width=False,
        )
        self.job.signal.register('generating_report', self.handle_generating_report)
        self.job.signal.register('generated_report', self.handle_generated_report)
        self.job.signal.register('running', lambda _: self._activity_indicator.enable())
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

    def handle_generating_report(self, video_filepath):
        # Display the path relative to the user-provided content path. Keep in mind that
        # `content_path` might be relative or have no parent directory.
        common_path = os.path.commonpath((self.job.content_path, video_filepath))
        relative_path = video_filepath[len(common_path):].strip(os.sep)
        self._activity_indicator.format = f'{{indicator}} {relative_path}'

    def handle_generated_report(self, video_filepath, report):
        # Remove file path from activity indicator in case we are waiting for the user to select
        # more playlists.
        self._activity_indicator.format = '{indicator}'

    @functools.cached_property
    def runtime_widget(self):
        return self._activity_indicator



import os
import re

# Regular expressions are matched against file paths.

# Path separator must be escaped in case we're running on Windows.
sep = re.escape(os.sep)

checksums = r'\.(?i:sfv|md5)$'

extras = (
    r'(?i:'
    # Directory named "extras".
    rf'{sep}extras{sep}'
    r'|'
    # Anything with the word "extras" with another word in front of it. This
    # should exclude the show "Extras".
    rf'{sep}.+[\. ]extras[\. ]'
    r'|'
    # Numbered extras (e.g. "Foo.S01.Extra1.mkv", "Foo.S01.Extra.2.mkv", .etc)
    rf'{sep}.+[\. ]extra[\. ]?\d+[\.]'
    r')'
)

images = r'\.(?i:png|jpg|jpeg)$'

nfo = r'\.(?i:nfo)$'

samples = (
    r'(?i:'
    # Sample directory
    rf'{sep}[!_0-]?sample{sep}'
    r'|'
    # Sample file name starts with release name
    rf'[^{sep}][\.\-_ ]sample\.mkv'
    r'|'
    # Sample file name ends with release name
    rf'{sep}sample[\!\-_].+\.mkv'
    r'|'
    # Sample file name starts with release name and ends with "sample-RLSGRP.mkv"
    r'[\.\-_!]?sample-[a-zA-Z0-9]+\.mkv'
    r'|'
    # Sample file name starts with "<characters that top-sort>sample"
    rf'{sep}[!#$%&*+\-\.]?sample\.mkv'
    r')'
)

subtitles = r'\.(?i:srt|idx|sub)$'



"""
Concrete :class:`~.TrackerConfigBase` subclass for NBL
"""

import base64

from ... import utils
from .. import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class NblTrackerConfig(base.TrackerConfigBase):
    defaults = {
        'anonymous': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description='Whether your username is displayed on your uploads.',
        ),
        'upload_url': base64.b64decode('aHR0cHM6Ly9uZWJ1bGFuY2UuaW8vdXBsb2FkLnBocA==').decode('ascii'),
        'announce_url': utils.configfiles.config_value(
            value='',
            description=(
                'The complete announce URL with your private passkey.\n'
                'Get it from the website: Shows -> Upload -> Your personal announce URL'
            ),
        ),
        'apikey': utils.configfiles.config_value(
            value='',
            description=(
                'Your personal private API key.\n'
                'Get it from the website: <USERNAME> -> Settings -> API keys'
            ),
        ),
        'exclude': (),
    }

    argument_definitions = {
        'submit': {
            ('--anonymous', '--an'): {
                'help': 'Hide your username for this submission',
                'action': 'store_true',
                # This must be `None` so it doesn't override the "anonymous"
                # value from the config file. See CommandBase.get_options().
                'default': None,
            },
            ('--tvmaze', '--tv'): {
                'help': 'TVmaze ID or URL',
                'type': utils.argtypes.webdb_id('tvmaze'),
            },
        },
    }



"""
Manage a bunch of jobs
"""

import collections

import logging  # isort:skip
_log = logging.getLogger(__name__)


class JobRunner(collections.abc.Mapping):
    """
    Wrapper around a bunch of :class:`~.JobBase` subclasses to manage them collectively

    This class subclasses :class:`~.collections.abc.Mapping` for easy access to jobs. Keys are job
    :attr:`~.JobBase.name`\\ s and values are the :class:`~.JobBase` instances.

    :param jobs: :class:`collections.abc.Iterable` of :class:`~.JobBase` subclasses
    :param str id: Any identifier (only used for debugging)
    """

    def __init__(self, jobs=(), id='anonymous'):
        self._id = str(id)
        self._jobs = {}
        self.add(*jobs)

    def __getitem__(self, job_name):
        return self._jobs[job_name]

    def __iter__(self):
        return iter(self._jobs)

    def __len__(self):
        return len(self._jobs)

    def add(self, *jobs):
        """
        Add `jobs`

        Every job must have a unique :class:`~.JobBase.name`.

        :param jobs: :class:`~.JobBase` subclasses
        """
        for job in jobs:
            if job.name in self._jobs:
                if job is not self._jobs[job.name]:
                    raise RuntimeError(f'Conflicting job name: {job.name!r}')
            else:
                _log.debug('%s: Adding job: %r', self._id, job)
                self._jobs[job.name] = job

    @property
    def all_jobs(self):
        """All :meth:`add`\\ ed jobs as a flat sequence"""
        return tuple(self._jobs.values())

    @property
    def enabled_jobs(self):
        """Same as :attr:`all_jobs`, but without any jobs that are not :attr:`~.JobBase.is_enabled`"""
        return tuple(job for job in self._jobs.values() if job.is_enabled)

    def start_more_jobs(self):
        """
        :meth:`~.JobBase.start` all jobs that are ready to start

        .. note:: This method must be called while an asyncio event loop is running because
            JobBase.start() calls JobBase.add_task(), and a task must be added to a loop to run.
        """
        for job in self.all_jobs:
            if (
                    not job.is_started
                    and not job.is_terminated
                    and job.autostart
                    and job.is_enabled
            ):
                _log.debug('%s: Starting job: %r', self._id, job)
                job.start()

    async def wait(self):
        """
        Wait for all jobs that are started and not finished

        .. note:: This does not raise exceptions from jobs. See :attr:`exceptions` after calling
            this method.
        """
        def get_running_jobs():
            return tuple(
                job
                for job in self.all_jobs
                if job.is_started and not job.is_finished
            )

        while running_jobs := get_running_jobs():
            _log.debug('%s: Waiting for running jobs:', self._id)
            for job in running_jobs:
                _log.debug('%s:    * %r', self._id, job)
            for job in running_jobs:
                await job.wait_finished()
                _log.debug('%s: Job finished: %r', self._id, job)
        _log.debug('%s: All running jobs finished', self._id)

    def terminate(self, reason=None):
        """Terminate :attr:`all_jobs`"""
        _log.debug('%s: Terminating all jobs', self._id)
        for job in self.all_jobs:
            job.terminate(reason=reason)

    @property
    def all_jobs_finished(self):
        """Whether all :attr:`enabled_jobs` are finished"""
        return all(job.is_finished for job in self.enabled_jobs)

    @property
    def exceptions(self):
        """Sequence of :class:`Exception` instances :attr:`~.JobBase.raised` by any job"""
        return tuple(
            job.raised
            for job in self.all_jobs
            if job.raised
        )



import collections
import functools
import json
import os
import re

from ... import errors, utils
from . import audio, text, video

import logging  # isort:skip
_log = logging.getLogger(__name__)

NO_DEFAULT_VALUE = object()

if utils.os_family() == 'windows':
    _mediainfo_executable = 'mediainfo.exe'
    _ffprobe_executable = 'ffprobe.exe'
else:
    _mediainfo_executable = 'mediainfo'
    _ffprobe_executable = 'ffprobe'


@functools.cache
def _run_mediainfo(path, *args):
    # Mediainfo can read .IFO and .MPLS files, which is useful for detecting subtitles. But those
    # are not found by find_main_video() so we have to make an exception here.
    if (
            not os.path.isdir(path)
            and utils.fs.file_extension(path).lower() in ('ifo', 'mpls')
    ):
        video_path = path
    else:
        video_path = utils.fs.find_main_video(path)

    # It's easier to check for readability than to interpret error output from mediainfo.
    utils.fs.assert_file_readable(video_path)
    cmd = (_mediainfo_executable, video_path, *args)

    # Translate DependencyError to ContentError so callers have to expect less exceptions. Do not
    # catch ProcessError because things like wrong mediainfo arguments are bugs.
    try:
        return utils.subproc.run(cmd)
    except errors.DependencyError as e:
        raise errors.ContentError(e) from e


def get_mediainfo_report(path):
    """
    ``mediainfo`` output as a string

    The parent directory of `path` is redacted.

    :param str path: Path to video file or directory

        For directories, the return value of :func:`find_main_video` is used.

    :raise ContentError: if anything goes wrong
    """
    def remove_parent_directory(mi, parent_directory=utils.fs.dirname(path)):
        # Remove parent directory, but only if it's not empty.
        if parent_directory:
            return mi.replace(parent_directory + os.sep, '')
        else:
            return mi

    def ensure_unique_id_exists(mi):
        # Some old files don't contain a unique ID and some mediainfo parsers cannot deal with that,
        # so we default to a fake "Unique ID" field.
        unique_id_regex = re.compile(r'^Unique ID\s*:\s*\d+', flags=re.MULTILINE)
        unique_id_default = 'Unique ID                                : 0 (0x0)'
        if unique_id_regex.search(mi):
            return mi
        else:
            assert mi.startswith('General\n')
            return (
                mi[:len('General\n')]
                + unique_id_default + '\n'
                + mi[len('General\n'):]
            )

    report = _run_mediainfo(path)
    return remove_parent_directory(ensure_unique_id_exists(report))


def get_tracks(path, default=NO_DEFAULT_VALUE):
    """
    ``mediainfo --Output=JSON`` as dictionary that maps each track's ``@type`` to a list of
    track dictionaries

    :param str path: Path to video file or directory

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    stdout = _run_mediainfo(path, '--Output=JSON')
    tracks = {}
    try:
        for track in json.loads(stdout)['media']['track']:
            if track['@type'] not in tracks:
                tracks[track['@type']] = []
            tracks[track['@type']].append(track)
    except (ValueError, TypeError) as e:
        raise RuntimeError(f'{path}: Unexpected mediainfo output: {stdout}: {e}') from e
    except KeyError as e:
        raise RuntimeError(f'{path}: Unexpected mediainfo output: {stdout}: Missing field: {e}') from e
    else:
        return tracks


def _get_default_track(tracks):
    # Find first track that is marked as default.
    for track in tracks:
        if track.get('Default') == 'Yes':
            return track

    # Default to first track.
    return tracks[0]


def lookup(path, keys, type=None, default=NO_DEFAULT_VALUE):
    """
    Return nested value from :func:`get_tracks`

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param keys: Iterable of nested keys/indexes in the return value of :func:`get_tracks`

    :param type: Callable that takes a :class:`str` value from the tracks returned by
        :func:`get_tracks` and returns a desired type (e.g. :class:`int`)

    :param default: Return value if `path` doesn't exist, raise :exc:`ValueError` if not provided

    For example, `("Audio", 0, "Language")` returns the language of the first audio track. If no
    language is defined return `default` or raise :exc:`~.ContentError` if `default` is not
    provided.

    In lists, instead of an index, you can use ``"DEFAULT"`` to find the default track in that list,
    e.g. `("Audio", "DEFAULT", "Language")` returns the language of the default audio track. The
    default track is either marked specifically as such in the container or, if no track is marked,
    the first track is used.

    Any exception from `type` will bubble up and must be caught by the caller.

    :raise ContentError: if `path` does not exist or if `keys` does not resolve to a value and
        `default` is not provided.
    """
    try:
        initial_value = value = get_tracks(path)
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        try:
            for key in keys:
                if isinstance(value, collections.abc.Mapping):
                    value = value[key]

                elif isinstance(value, collections.abc.Sequence) and not isinstance(value, str):
                    if key == 'DEFAULT':
                        value = _get_default_track(value)
                    else:
                        value = value[key]

                else:
                    value = default

        except (KeyError, IndexError, TypeError):
            value = default

        if value is NO_DEFAULT_VALUE:
            raise errors.ContentError(f'Unable to lookup {keys!r} in {initial_value!r}')
        elif type is not None:
            return type(value)
        else:
            return value


def get_duration(path, default=NO_DEFAULT_VALUE):
    """
    Return video duration in seconds (float)

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        return get_duration_from_ffprobe(path, default=default)
    except (errors.DependencyError, errors.ProcessError):
        return get_duration_from_mediainfo(path, default=default)


def get_duration_from_ffprobe(path, default=NO_DEFAULT_VALUE):
    """
    Return video duration in seconds (float) from ``ffprobe``

    :param str path: Path to video file or directory

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    main_video = utils.fs.find_main_video(path)
    cmd = (
        _ffprobe_executable,
        '-v', 'error', '-show_entries', 'format=duration',
        '-of', 'default=noprint_wrappers=1:nokey=1',
        f'file:{main_video}',
    )
    length = utils.subproc.run(cmd, ignore_errors=True)
    try:
        return float(length.strip())
    except ValueError as e:
        raise RuntimeError(f'Unexpected output from {cmd}: {length!r}') from e


def get_duration_from_mediainfo(path, default=NO_DEFAULT_VALUE):
    """
    Return video duration in seconds (float) from ``mediainfo``

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        all_tracks = get_tracks(path)
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        try:
            return float(all_tracks['General'][0]['Duration'])
        except (KeyError, IndexError, ValueError, TypeError) as e:
            raise RuntimeError(f'Unexpected tracks from {path}: {all_tracks!r}') from e



"""
Wrapper for ``mediainfo`` command
"""

import os

from .. import errors, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MediainfoJob(base.JobBase):
    """
    Get output from ``mediainfo`` command

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``generating_report``
            Emitted before mediainfo report generation begins. Registered callbacks get the video
            file path ``mediainfo`` was called with.

        ``generated_report``
            Emitted after mediainfo report generation has finished. Registered callbacks get the
            video file path ``mediainfo`` was called with and the report text.
    """

    name = 'mediainfo'
    label = 'Mediainfo'

    _DEFAULT_FORMAT = '{MEDIAINFO}'

    # Caching is done on a per-file basis in the daemon process.
    cache_id = None

    @property
    def content_path(self):
        """Instantiation argument of the same name"""
        return self._content_path

    def initialize(
            self,
            *,
            content_path,
            from_all_videos=False,
            exclude_files=(),
            format='{MEDIAINFO}',
    ):
        """
        Set internal state

        :param content_path: Path to video file or directory that contains a video file

        :param bool from_all_videos: Whether to get ``mediainfo`` output from each video file or
            only from the first video

        :param exclude_files: Sequence of glob patterns (:class:`str`) and :class:`re.Pattern`
            objects (return value from :func:`re.compile`) that are matched against each relative
            path beneath `content_path`

            Glob patterns are matched case-insensitively.

            .. note:: Non-video files and stuff like `Sample.mkv` are always
                      excluded (see :func:`.fs.find_main_videos`).

        :param format: String that contains the placeholder ``"{MEDIAINFO}"``,
            which is replaced by the actual mediainfo

            Any other placeholders are ignored.
        """
        self._content_path = content_path
        self._from_all_videos = from_all_videos
        self._exclude_files = exclude_files
        self._format = format
        self._reports_by_file = {}
        self._mediainfo_process = None
        self.signal.add('generating_report')
        self.signal.add('generated_report')
        self.signal.register('generated_report', self._store_report_by_file)
        self.signal.register('finished', self._hide_job)

    def _hide_job(self, job):
        # Do not display the mediainfo report as output. It can be very long.
        self.hidden = True

    async def run(self):
        self._mediainfo_process = utils.daemon.DaemonProcess(
            target=_mediainfo_process,
            kwargs={
                'from_all_videos': self._from_all_videos,
                'exclude_files': self._exclude_files,
                'cache_directory': self.cache_directory,
                'ignore_cache': self.ignore_cache,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._mediainfo_process.start()

        if utils.disc.is_disc(self._content_path, multidisc=True):
            await self._process_playlists()
        else:
            await self._process_file_or_directory()

        await self._mediainfo_process.join()
        self._add_mediainfo_reports()

    def terminate(self, reason=None):
        if self._mediainfo_process:
            self._mediainfo_process.stop()
        super().terminate(reason=reason)

    async def _process_playlists(self):
        # Receive selected playlists from PlaylistsJob and pass them on to _mediainfo_process().
        async for _discpath, playlists in self.receive_all('playlists', 'playlists_selected', only_posargs=True):
            _log.debug('Sending playlists to _mediainfo_process(): %r', playlists)
            for playlist in playlists:
                self._mediainfo_process.send(utils.daemon.MsgType.info, {'source': playlist})
        _log.debug('All playlists received')

        # Tell _mediainfo_process() that there won't be any more playlists.
        self._mediainfo_process.send(utils.daemon.MsgType.info, {'source': _ALL_SOURCES_SELECTED})

    async def _process_file_or_directory(self):
        self._mediainfo_process.send(utils.daemon.MsgType.info, {'source': self._content_path})
        self._mediainfo_process.send(utils.daemon.MsgType.info, {'source': _ALL_SOURCES_SELECTED})

    @base.unless_job_is_finished
    def _handle_info(self, info):
        if 'generating_report' in info:
            self.signal.emit('generating_report', info['generating_report'])
        elif 'generated_report' in info:
            self.signal.emit('generated_report', info['video_filepath'], info['generated_report'])
        else:
            raise RuntimeError(f'Unexpected info: {info!r}')

    @base.unless_job_is_finished
    def _handle_error(self, error):
        if isinstance(error, errors.ContentError):
            self.error(error)
        else:
            self.exception(error)

    def _add_mediainfo_reports(self):
        for report in self.reports_by_file.values():
            # We are not using self._format.format(MEDIAINFO=report) because that turns "{" and "}"
            # into special characters that must be escaped, which the user should not expect.
            self.add_output(self._format.replace('{MEDIAINFO}', report))

    def _store_report_by_file(self, video_filepath, mediainfo):
        _log.debug(f'Storing mediainfo report for {video_filepath}')
        self._reports_by_file[video_filepath] = mediainfo

    @property
    def reports_by_file(self):
        """
        Map video file paths to ``mediainfo`` outputs gathered so far

        .. note:: For VIDEO_TS releases, one mediainfo is made for an ``.IFO`` file and a second
                  mediainfo is made for a ``.VOB`` file.
        """
        return self._reports_by_file.copy()


# NOTE: This cannot be the typical `object()` constant because its identity differs between the main
# process and _mediainfo_process() and `... is _ALL_SOURCES_SELECTED` doesn't work as
# expected.
_ALL_SOURCES_SELECTED = '_ALL_SOURCES_SELECTED'

def _mediainfo_process(output_queue, input_queue, *, from_all_videos, exclude_files, cache_directory, ignore_cache):
    # This process reads user-selected video sources from the main process and runs `mediainfo` on
    # them.
    while True:
        source = utils.daemon.read_input_queue_key(input_queue, 'source')
        utils.daemon.maybe_terminate(input_queue)
        if source == _ALL_SOURCES_SELECTED:
            # No more values will be sent on `input_queue`.
            break
        else:
            _send_mediainfo_reports_for_source(
                output_queue=output_queue,
                input_queue=input_queue,
                source=source,
                from_all_videos=from_all_videos,
                exclude_files=exclude_files,
                cache_directory=cache_directory,
                ignore_cache=ignore_cache,
            )


def _send_mediainfo_reports_for_source(
        output_queue, input_queue,
        *, source, from_all_videos, exclude_files, cache_directory, ignore_cache,
):
    for video_filepath in _get_filepaths_for_source(source, from_all_videos, exclude_files):
        utils.daemon.maybe_terminate(input_queue)

        output_queue.put((utils.daemon.MsgType.info, {'generating_report': video_filepath}))
        try:
            mediainfo_report = _get_mediainfo_report(video_filepath, cache_directory, ignore_cache)
        except errors.ContentError as e:
            output_queue.put((utils.daemon.MsgType.error, e))
        else:
            output_queue.put((utils.daemon.MsgType.info, {
                'generated_report': mediainfo_report,
                'video_filepath': video_filepath,
            }))


def _get_mediainfo_report(video_filepath, cache_directory, ignore_cache):
    cache_filepath = _get_cache_filepath(video_filepath, cache_directory)

    if not ignore_cache:
        # Try to read mediainfo from cache file or fail silently.
        try:
            with open(cache_filepath, 'r') as f:
                return f.read()
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            _log.debug(f'Failed to read cache file {cache_filepath}: {msg}')

    # Create mediainfo report.
    mediainfo_report = utils.mediainfo.get_mediainfo_report(video_filepath)

    # Write mediainfo report to cache or fail silently.
    try:
        with open(cache_filepath, 'w') as f:
            f.write(mediainfo_report)
    except OSError as e:
        msg = e.strerror if e.strerror else str(e)
        _log.debug(f'Failed to write cache file {cache_filepath}: {msg}')
    finally:
        return mediainfo_report


def _get_cache_filepath(video_filepath, cache_directory):
    return os.path.join(
        cache_directory,
        f'{utils.fs.basename(video_filepath)}.{utils.fs.file_size(video_filepath) or 0}.mediainfo',
    )


def _get_filepaths_for_source(source, from_all_videos, exclude_files):
    if isinstance(source, utils.disc.Playlist):
        if source.type == 'dvd':
            return (
                source.filepath,      # IFO
                source.largest_item,  # VOB
            )
        else:
            return (source.largest_item,)

    elif os.path.isdir(source):
        video_files = utils.fs.find_main_videos(source, exclude=exclude_files)
        if from_all_videos:
            return video_files
        else:
            return (video_files[0],)

    else:
        # Source is file.
        return (source,)



"""
Custom job that provide the result of a coroutine
"""

import collections

from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CustomJob(JobBase):
    """
    Wrapper around coroutine function

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``indicate_activity``
            Emitted by :meth:`indicate_activity`. Registered callbacks get the `indicate_activity`
            and `text` arguments as positional arguments.
    """

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    def initialize(self, *, name, label, worker, catch=()):
        """
        Set internal state

        :param name: Name for internal use

        :param label: Name for user-facing use

        :param worker: Coroutine function that takes the job instance as a positional argument and
            returns this job's output

            If `worker` returns any iterable that isn't a string, each item is passed to
            :attr:`add_output`.

            If `worker` returns `None`, it is interpreted as "no output" and the job fails. An empty
            string is valid output and does not mean the job failed.

        :param catch: Sequence of :class:`Exception` classes to catch from `worker` and pass to
            :meth:`~.JobBase.error`

        This job finishes when `worker` returns.

        :class:`~.asyncio.CancelledError` from `worker` is ignored. Any other
        exceptions are passed to :meth:`~.JobBase.exception`.
        """
        self._name = str(name)
        self._label = str(label)
        self._worker = worker
        self._expected_exceptions = tuple(catch)
        self.signal.add('indicate_activity')

    def indicate_activity(self, indicate_activity=None, *, text=None):
        """
        Emit the ``indicate_activity`` :attr:`~.JobBase.signal`

        :param indicate_activity: `True`, `False` or `None` (i.e. do not change the current state)
        :param text: :class:`str` or `None` (i.e. do not change the current state)
        """
        self.signal.emit('indicate_activity', indicate_activity, text)

    async def run(self):
        try:
            result = await self._worker(self)
        except self._expected_exceptions as e:
            self.error(e)
        else:
            self._handle_worker_result(result)

    def _handle_worker_result(self, result):
        if result is not None:
            if (
                    not isinstance(result, str)
                    and isinstance(result, collections.abc.Iterable)
            ):
                # Worker returned list/tuple/etc
                for r in result:
                    self.add_output(r)
            else:
                self.add_output(result)



"""
Concrete :class:`~.TrackerConfigBase` subclass for PTP
"""

import base64

from ... import utils
from ..base import TrackerConfigBase, exclude
from . import metadata

PtpImageHost = utils.types.ImageHost(
    disallowed=(
        ()
        if utils.is_running_in_development_environment() else
        ('dummy',)
    ),
)


class PtpReleaseType(str):
    def __new__(cls, text):
        for type, regex in metadata.types.items():
            if regex.search(text):
                return super().__new__(cls, type)
        raise utils.argtypes.ArgumentTypeError(f'Invalid type: {text}')


class PtpTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii'),
        'username': '',
        'password': '',
        'cookies_filepath': utils.configfiles.config_value(
            value='',
            description=(
                'File that stores permanent session cookies.\n'
                'If this is not set, a new user session is started for each upload.'
            ),
        ),
        'announce_url': utils.configfiles.config_value(
            value='',
            description='Your personal announce URL.',
        ),
        'image_host': utils.configfiles.config_value(
            value=utils.types.ListOf(
                item_type=PtpImageHost,
                default=('ptpimg', 'freeimage'),
                separator=',',
            ),
            description=(
                'List of image hosting service names. The first service is normally used '
                'with the others as backup if uploading to the first fails.\n'
                + 'Supported services: ' + ', '.join(PtpImageHost.options)
            ),
        ),
        'screenshots_from_movie': utils.configfiles.config_value(
            value=utils.types.Integer(3, min=3, max=10),
            description='How many screenshots to make for single-video uploads.',
        ),
        'screenshots_from_episode': utils.configfiles.config_value(
            value=utils.types.Integer(2, min=2, max=10),
            description='How many screenshots to make per video for multi-video uploads.',
        ),
        'exclude': (
            exclude.checksums,
            exclude.images,
            exclude.nfo,
            exclude.samples,
        ),
    }

    argument_definitions = {
        'submit': {
            ('--hardcoded-subtitles', '--hs'): {
                'help': 'Video contains hardcoded subtitles',
                'action': 'store_true',
            },
            ('--no-english-subtitles', '--nes'): {
                'help': (
                    'Whether release contains no English audio and no English subtitles.'
                    '\n'
                    'This is autodetected reliably if all audio and subtitle tracks have '
                    'a correct language tag. If not, you are asked interactively.\n'
                    'Subtitle languages are detected in *.idx/sub, VIDEO_TS trees '
                    'and *.srt/ssa/ass/vtt by language code in the file name, e.g. '
                    '"Foo.en.srt".'
                ),
                'metavar': 'BOOL',
                'type': utils.argtypes.bool_or_none,
            },
            ('--not-main-movie', '--nmm'): {
                'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
                'action': 'store_true',
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--poster', '--po'): {
                'help': 'Path or URL to movie poster',
            },
            ('--nfo',): {
                'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
            },
            ('--imdb', '--im'): {
                'help': 'IMDb ID or URL',
                'type': utils.argtypes.webdb_id('imdb'),
            },
            ('--source', '--so'): {
                'help': (
                    'Original source of this release\n'
                    + 'Should vaguely match: '
                    + ', '.join(metadata.sources)
                ),
            },
            ('--type', '--ty'): {
                'help': (
                    'General category of this release\n'
                    + 'Must vaguely match: '
                    + ', '.join(metadata.types)
                ),
                'type': PtpReleaseType,
            },
            ('--screenshots-count', '--ssc'): {
                'help': 'How many screenshots to make per video file',
                'type': utils.argtypes.number_of_screenshots(min=1, max=10),
            },
            ('--screenshots', '--ss'): {
                'help': (
                    'Path(s) to existing screenshot file(s)\n'
                    'Directories are searched recursively.\n'
                    'Precreated screenshots are used in addition to automatically generated screenshots.'
                ),
                'nargs': '+',
                'action': 'extend',
                'type': utils.argtypes.files_with_extension('png'),
                'metavar': 'SCREENSHOT',
            },
            ('--upload-token', '--ut'): {
                'help': 'Upload token from staff',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



"""
Interactive text user interface and job manager
"""

import asyncio

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import HSplit, Window, to_container
from prompt_toolkit.output import create_output

from ...jobs import JobRunner
from . import jobwidgets, style

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TUI:
    def __init__(self):
        self._jobrunner = JobRunner(id='TUI')
        self._widgets = {}     # Map job names to JobWidgetBase instances.
        self._containers = {}  # Map job names to prompttoolkit.Container instances.
        self._focused_job_name = None
        self._app = self._make_app()
        self._loop = asyncio.get_event_loop()
        self._unhandled_exception = None
        self._loop.set_exception_handler(self._handle_exception)

    def _handle_exception(self, loop, context):
        exception = context.get('exception')
        if exception:
            _log.debug('Caught unhandled exception: %r', exception)
            _log.debug('Unhandled exception context: %r', context)
            if not self._unhandled_exception:
                self._unhandled_exception = exception
                self._jobrunner.terminate(reason=f'Unhandled exception: {exception!r}')

    def _make_app(self):
        self._jobs_container = HSplit(
            # FIXME: Layout does not accept an empty list of children, so we add an initial empty
            #        Window that doesn't display anything.
            #        https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1257
            children=[Window()],
            style='class:default',
        )
        self._layout = Layout(self._jobs_container)

        kb = KeyBindings()

        @kb.add('escape')
        @kb.add('c-g')
        @kb.add('c-q')
        @kb.add('c-c')
        def _(_event, self=self):
            self._jobrunner.terminate(reason='User terminated application')

        @kb.add('escape', 'I')
        def _(_event, self=self):
            _log.debug('=== CURRENT JOBS ===')
            for job in self._jobrunner.all_jobs:
                _log.debug(' %s (%d tasks):', job, len(job._tasks))
                for task in job._tasks:
                    _log.debug('   %r', task)

            _log.debug('Focused job: %r: %r', self._focused_job_name, self._focused_job)
            _log.debug('Focused widget [layout ]: %r', self._layout.current_control)
            _log.debug('Focused widget [focused]: %r', self._widgets.get(self._focused_job_name, None))
            _log.debug('Focused container: %r', self._containers.get(self._focused_job_name, None))
            _log.debug('Layout.has_focus(%r): %r', self._layout.current_control,
                       self._layout.has_focus(self._layout.current_control))

        return Application(
            # Write TUI to stderr if stdout is redirected. This is useful for allowing the user to
            # make decisions in the TUI (e.g. selecting an item from search results) while
            # redirecting the final output (e.g. an IMDb ID).
            output=create_output(always_prefer_tty=True),
            layout=self._layout,
            key_bindings=kb,
            style=style.style,
            full_screen=False,
            erase_when_done=False,
            mouse_support=False,
            # Determine the currently active job *after* Application was invalidated. At the time of
            # this writing, this is important so we don't try to focus widgets from a job that has
            # just finished, which can result in RuntimeErrors because signals cannot be emitted on
            # finished jobs.
            before_render=self._update_jobs_container,
        )

    def _add_jobs(self, jobs):
        for job in jobs:
            self._add_job(job)

        # Add job widgets to the main container widget.
        self._update_jobs_container()

        # Register signal callbacks. It's probably best to do this after all jobs were added so that
        # signals aren't emitted before all jobs are available.
        self._connect_jobs(jobs)

    def _add_job(self, job):
        self._jobrunner.add(job)
        self._widgets[job.name] = jobwidgets.JobWidget(job, self._app)
        _log.debug('Job widget: %r: widget=%r', job.name, self._widgets[job.name])
        self._containers[job.name] = to_container(self._widgets[job.name])
        _log.debug('Job widget: %r: container=%r', job.name, self._containers[job.name])

    def _connect_jobs(self, jobs):
        for job in jobs:
            # Every time a job finishes, other jobs can become enabled due to the dependencies on
            # other jobs or other conditions. We also want to display the next interactive job when
            # an interactive job is done.
            job.signal.register('finished', self._handle_job_finished)

            # A job can also signal explicitly that we should update the job widgets, e.g. to start
            # previously disabled jobs.
            job.signal.register('refresh_ui', self._refresh)

    def _handle_job_finished(self, finished_job):
        assert finished_job.is_finished, f'{finished_job.name} is actually not finished'

        # Start enabled but not yet started jobs and display the next interactive job. This also
        # generates the regular output if output from all jobs was read from cache and the TUI exits
        # immediately.
        self._refresh()

        # Terminate all jobs and exit if job finished with non-zero exit code.
        if finished_job.exit_code != 0:
            self._jobrunner.terminate(reason=f'Job failed: {finished_job.name}: {finished_job.raised!r}')

    def _refresh(self):
        self._jobrunner.start_more_jobs()
        self._update_jobs_container()
        self._app.invalidate()

    @property
    def _focused_job(self):
        if self._focused_job_name:
            return self._jobrunner[self._focused_job_name]

    # We accept one argument because `before_render` calls this method with the Application
    # instance.
    def _update_jobs_container(self, _=None):
        # Unfocus focused job if it is finished.
        if self._focused_job and self._focused_job.is_finished:
            # _log.debug('UPDATE JOB CONTAINER: Unfocusing: %r', self._focused_job_name)
            self._focused_job_name = None

        enabled_jobs = self._jobrunner.enabled_jobs
        # _log.debug('UPDATE JOB CONTAINER: Enabled jobs: %r', [job.name for job in enabled_jobs])

        # Don't change focus if we already have a focused job. If another job becomes interactive
        # asynchronously (e.g. because a background job finished), it must not steal focus from the
        # currently focused job.
        if not self._focused_job_name:
            # Focus next interactive job.
            for job in enabled_jobs:
                if (
                        job.is_started
                        and not job.is_finished
                        and self._widgets[job.name].is_interactive
                ):
                    # _log.debug('UPDATE JOB CONTAINER: Focusing next interactive job: %s', job.name)
                    self._focused_job_name = job.name
                    break
        #     else:
        #         _log.debug('UPDATE JOB CONTAINER: No focusable widget found: %r', [
        #             {
        #                 'name': job.name,
        #                 'is_started': job.is_started,
        #                 'is_finished': job.is_finished,
        #                 'is_interactive': self._widgets[job.name].is_interactive,
        #             }
        #             for job in enabled_jobs
        #         ])
        # else:
        #     _log.debug('UPDATE JOB CONTAINER: Preserving focus: %r', self._focused_job_name)

        # Display focused job, finished jobs and all background jobs.
        self._jobs_container.children[:] = (
            self._containers[job.name]
            for job in enabled_jobs
            if (
                job.name == self._focused_job_name
                or job.is_finished
                or not self._widgets[job.name].is_interactive
            )
        )

        if self._focused_job_name:
            # Actually focus the focused job.
            try:
                self._layout.focus(self._containers[self._focused_job_name])
            except ValueError:
                # A job may hardcode `is_interactive = True` even though it currently is not
                # focusable. This happens, for example, if the job is still autodetecting before the
                # user can fix or confirm the autodetected value. In that case, we can either wait
                # for the job to become focusable or focus another interactive job in the meantime.
                pass

    def run(self, jobs):
        """
        Synchronous wrapper around :meth:`run_async`
        """
        return self._loop.run_until_complete(self.run_async(jobs))

    async def run_async(self, jobs):
        """
        Block while `jobs` are running and the user interface is up

        :param jobs: Iterable of :class:`~.jobs.base.JobBase` instances

        :raise: any :class:`Exception` from any job or the user interface

        :return: :attr:`~.JobBase.exit_code` from the first failed job or ``0`` if all jobs
            succeeded
        """
        # Run the TUI in a background task that can be cancelled. This makes it easier to terminate
        # self._app.run_async() and get an exception from it.
        self._app_task = self._loop.create_task(
            self._app.run_async(set_exception_handler=False)
        )
        _log.debug('============ TUI is now running ============')

        # Add jobs to our JobRunner and create TUI widgets for them.
        self._add_jobs(jobs)

        # Start initially enabled jobs. This must be done asynchronously because we need a running
        # asyncio event loop for JobBase.add_task(), which is called by JobBase.start().
        self._jobrunner.start_more_jobs()

        # Wait for jobs to finish.
        await self._jobrunner.wait()
        _log.debug('All jobs finished:')
        for job in self._jobrunner.all_jobs:
            _log.debug(' * %r', job)

        # Wait for application task.
        self._app_task.cancel()
        try:
            await self._app_task
        except asyncio.CancelledError:
            pass
            # self._jobrunner.terminate(reason=f'Application was cancelled: {self._app_task!r}')
        _log.debug('============ TUI has stopped running ============')

        # Raise any stored exception.
        self._maybe_raise_exception()

        # Return application exit code (e.g. 0=success, 1=failure).
        return self._get_exit_code()

    def _maybe_raise_exception(self):
        if self._unhandled_exception:
            # Some task raised an exception that wasn't handled properly.
            _log.debug('Raising unhandled exception: %r', self._unhandled_exception)
            raise self._unhandled_exception

        if self._jobrunner.exceptions:
            # One or more jobs raised an exception via JobBase.exception().
            _log.debug('Raising first job exception: %r', self._jobrunner.exceptions)
            raise self._jobrunner.exceptions[0]

    def _get_exit_code(self):
        for job in self._jobrunner.all_jobs:
            _log.debug('Exit code of %r: %r', job.name, job.exit_code)

        # First non-zero exit_code is the application exit_code
        for job in self._jobrunner.all_jobs:
            if job.exit_code not in (0, None):
                _log.debug('Exiting with exit code from %s: %r', job.name, job.exit_code)
                return job.exit_code

        return 0



"""
Create screenshots from video file(s)
"""

import collections
import inspect
import os

from .. import errors, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)

shutil = utils.LazyModule(module='shutil', namespace=globals())


DEFAULT_NUMBER_OF_SCREENSHOTS = 2


class ScreenshotsJob(base.JobBase):
    r"""
    Create screenshots from video file(s) or :class:`~.Playlist` instances

    To get :class:`~.Playlist`\ s, this job relies on a :class:`~.PlaylistsJob` instance to be
    available via a job :attr:`~.JobBase.siblings` named "playlists".

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``screenshots_total``
            Emitted before screenshots are created. Registered callbacks get the
            total number of screenshots as a positional argument.
    """

    name = 'screenshots'
    label = 'Screenshots'

    # Caching of original/optimized screenshots is handled in the _screenshots/optimize_process()
    # functions based on file names, which we generate and should be unique.
    cache_id = None

    def initialize(self, *, content_path, precreated=(),
                   exclude_files=(), timestamps=(), count=0,
                   from_all_videos=False, optimize='default'):
        r"""
        Set internal state

        :param str content_path: Path to file or directory or sequence of paths

        :param precreated: Sequence of paths of already existing screenshots

            These do not count towards the wanted number of screenshots.
            `count` screenshots are created in addition to any precreated
            screenshots.

        :param exclude_files: Sequence of glob patterns (:class:`str`) and
            :class:`re.Pattern` objects (return value from :func:`re.compile`)
            that are matched against the relative path beneath each `source`

            Glob patterns are matched case-insensitively.

            .. note:: Non-video files and stuff like `Sample.mkv` are always
                      excluded by :func:`.fs.find_main_videos`.

        :param timestamps: Screenshot positions in the video
        :type timestamps: sequence of "[[H+:]M+:]S+" strings or seconds

        :param count: Number of screenshots to make or :func:`callable` that returns how many
            screenshots to make

            If this is a callable (synchronous or asynchronous), it is called when this job is
            :meth:`~.JobBase.start`\ ed, and :attr:`count` is ``0`` until then.

        :param bool from_all_videos: Whether to take `count` screenshots from
            each video file or only from the first video

            See :func:`.fs.find_main_videos` for more information.

        :param optimize: `level` argument for :func:`~image.optimize`

            If this is ``"default"``, missing optimization dependencies are
            silently ignored.

        If `timestamps` and `count` are not given, screenshot positions are
        picked at even intervals. If `count` is larger than the number of
        `timestamps`, more timestamps are added.
        """
        self._content_path = content_path
        self._precreated = precreated
        self._exclude_files = exclude_files
        self._timestamps = timestamps
        if callable(count):
            self._count = 0
            self._count_callable = count
        else:
            self._count = count
            self._count_callable = None
        self._from_all_videos = from_all_videos
        self._optimize = optimize
        self._screenshots_created = 0
        self._screenshots_total = -1
        self._screenshots_process = None
        self._optimize_process = None
        self._screenshots_by_file = collections.defaultdict(list)
        self.signal.add('screenshots_total', record=True)

    async def run(self):
        """Execute subprocesses for screenshot creation and optimization"""
        # Execute subprocesses
        await self._execute_screenshots_process()
        if self._optimize not in ('none', None):
            self._execute_optimize_process()

        if utils.disc.is_disc(self._content_path, multidisc=True):
            await self._process_playlists()
        else:
            await self._process_file_or_directory()

        # Wait for subprocesses
        await self._screenshots_process.join()
        if self._optimize_process:
            await self._optimize_process.join()

    async def _execute_screenshots_process(self):
        if inspect.iscoroutinefunction(self._count_callable):
            self._count = await self._count_callable()
        elif callable(self._count_callable):
            self._count = self._count_callable()
        _log.debug('Screenshots per video file: %s', self._count)

        self._screenshots_process = utils.daemon.DaemonProcess(
            name='_screenshots_process',
            target=_screenshots_process,
            kwargs={
                'precreated': self._precreated,
                'exclude_files': self._exclude_files,
                'timestamps': self._timestamps,
                'count': self._count,
                'from_all_videos': self._from_all_videos,
                'output_dir': self.cache_directory,
                'overwrite': self.ignore_cache,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._screenshots_process.start()

    def _execute_optimize_process(self):
        self._optimize_process = utils.daemon.DaemonProcess(
            name='_optimize_process',
            target=_optimize_process,
            kwargs={
                'level': self._optimize,
                'overwrite': self.ignore_cache,
                # Ignore missing dependecy if we do "default" optimization
                'ignore_dependency_error': self._optimize == 'default',
                'cache_directory': self.cache_directory,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._optimize_process.start()

    async def _process_playlists(self):
        async for _discpath, playlists in self.receive_all('playlists', 'playlists_selected', only_posargs=True):
            _log.debug('Sending playlists to _screenshots_process(): %r', playlists)
            for playlist in playlists:
                self._screenshots_process.send(utils.daemon.MsgType.info, {'source': playlist})
        self._screenshots_process.send(utils.daemon.MsgType.info, {'source': _ALL_SOURCES_SELECTED})

    async def _process_file_or_directory(self):
        self._screenshots_process.send(utils.daemon.MsgType.info, {'source': self._content_path})
        self._screenshots_process.send(utils.daemon.MsgType.info, {'source': _ALL_SOURCES_SELECTED})

    @base.unless_job_is_finished
    def _handle_info(self, info):
        if 'screenshots_total' in info:
            self._screenshots_total = info['screenshots_total']
            self.signal.emit('screenshots_total', self._screenshots_total)

        elif 'screenshot_filepath' in info:
            if self._optimize_process:
                _log.debug('Screenshot: %s: %.2f KiB',
                           info['screenshot_filepath'],
                           utils.fs.file_size(info['screenshot_filepath']) or 0 / 1024)
                self._optimize_process.send(utils.daemon.MsgType.info, info)

            else:
                _log.debug('Screenshot: %s: %.2f KiB',
                           info['screenshot_filepath'],
                           utils.fs.file_size(info['screenshot_filepath']) or 0 / 1024)
                self.add_output(info['screenshot_filepath'], info['video_filepath'], info['source'])

        elif 'optimized_screenshot_filepath' in info:
            _log.debug('Optimized %s: %.2f KiB',
                       info['optimized_screenshot_filepath'],
                       utils.fs.file_size(info['optimized_screenshot_filepath']) or 0 / 1024)
            self.add_output(info['optimized_screenshot_filepath'], info['video_filepath'], info['source'])

        else:
            raise RuntimeError(f'Unexpected info: {info!r}')

        if self._optimize_process and self.screenshots_created == self.screenshots_total:
            self._optimize_process.stop()

    @base.unless_job_is_finished
    def _handle_error(self, error):
        if (
                isinstance(error, (
                    errors.ScreenshotError,
                    errors.ImageOptimizeError,
                    errors.DependencyError,
                ))
                or not isinstance(error, BaseException)
        ):
            self.error(error)
        else:
            raise error

    def terminate(self, reason=None):
        """
        Stop screenshot creation and optimization subprocesses before
        terminating the job
        """
        if self._screenshots_process:
            self._screenshots_process.stop()
        if self._optimize_process:
            self._optimize_process.stop()
        super().terminate(reason=reason)

    @property
    def exit_code(self):
        """`0` if all screenshots were made, `1` otherwise, `None` if unfinished"""
        if self.is_finished:
            if self.screenshots_total < 0:
                # Job is finished but _screenshots_process() never told us home many screenshots
                # should be created. That means we're either using previously cached output or the
                # job was cancelled while _screenshots_process() was still initializing.
                if self.output:
                    # If we have cached output, assume the cached number of screenshots is what the
                    # user wanted because the output of unsuccessful jobs is not cached (see
                    # JobBase._write_cache()).
                    return 0
                else:
                    return 1
            elif len(self.output) == self.screenshots_total:
                return 0
            else:
                return 1

    @property
    def exclude_files(self):
        """
        Sequence of glob and :class:`regex <re.Pattern>` patterns to exclude

        See :meth:`initialize` for more information.

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._exclude_files

    @exclude_files.setter
    @base.raise_if_started
    def exclude_files(self, exclude_files):
        self._exclude_files = exclude_files

    @property
    def from_all_videos(self):
        """
        Whether to make screenshots from all video files or only the first

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._from_all_videos

    @from_all_videos.setter
    @base.raise_if_started
    def from_all_videos(self, from_all_videos):
        self._from_all_videos = from_all_videos

    @property
    def count(self):
        """
        How many screenshots to make per video file

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._count

    @count.setter
    @base.raise_if_started
    def count(self, count):
        self._count = count

    @property
    def timestamps(self):
        """
        Specific list of timestamps to make

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._timestamps

    @timestamps.setter
    @base.raise_if_started
    def timestamps(self, timestamps):
        self._timestamps = timestamps

    @property
    def screenshots_total(self):
        """
        Total number of screenshots to make

        .. note:: This is ``-1`` until the subprocess that creates the
                  screenshots is executed and determined the number of
                  screenshots.
        """
        return self._screenshots_total

    @property
    def screenshots_created(self):
        """Total number of screenshots made so far"""
        return self._screenshots_created

    @property
    def screenshots_by_file(self):
        """Map video file paths to sequences of generated screenshot file paths so far"""
        return {
            video_filepath: tuple(screenshot_paths)
            for video_filepath, screenshot_paths in self._screenshots_by_file.items()
        }

    def add_output(self, screenshot_filepath, video_filepath, source):
        """
        Populate :attr:`~.JobBase.output` and :attr:`screenshots_by_file` and bump
        :attr:`screenshots_created`

        :param screenshots_filepath: Path to screenshot file
        :param video_filepath: Path to video file `screenshot_filepath` is from
        :param source: Path to release directory, :class:`~.Playlist` instance or the same as
            `video_filepath`
        """

        def normalize_path(path):
            # Resolve symbolic links and make path absolute.
            return os.path.abspath(os.path.realpath(path))

        # Copy screenshot to home_directory if it is different from cache_directory.
        home_directory = normalize_path(self.home_directory)
        cache_directory = normalize_path(self.cache_directory)
        if home_directory != cache_directory:
            destination_path = self.home_directory or '.'
            try:
                screenshot_filepath = shutil.copy2(screenshot_filepath, destination_path)
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                self.error(f'{msg}: {destination_path}')
                return

        # Register successfully created screenshot internally.
        self._screenshots_created += 1
        if isinstance(source, utils.disc.Playlist):
            self._screenshots_by_file[source.filepath].append(screenshot_filepath)
        else:
            self._screenshots_by_file[video_filepath].append(screenshot_filepath)

        return super().add_output(screenshot_filepath)

def _screenshots_process(
        output_queue, input_queue,
        *,
        precreated, exclude_files, timestamps, count,
        from_all_videos, output_dir, overwrite,
):
    # First we have to collect all video files we want to create screenshots from. This allows us to
    # report the total number of screenshots.
    screenshot_infos = _get_screenshot_infos(
        input_queue=input_queue,
        custom_timestamps=timestamps,
        count=count,
        exclude_files=exclude_files,
        from_all_videos=from_all_videos,
    )

    # How many screenshots we are going to make. This is used for displaying progress.
    screenshots_total = len(precreated) + len(screenshot_infos)
    output_queue.put((utils.daemon.MsgType.info, {'screenshots_total': screenshots_total}))

    # Feed user-provided, precreated screenshots back into the normal processing pipeline.
    for screenshot_filepath in precreated:
        output_queue.put((utils.daemon.MsgType.info, {
            'screenshot_filepath': screenshot_filepath,
            'video_filepath': '',
            'source': '',
        }))

    try:
        for source, video_filepath, screenshot_filename_base, timestamp in screenshot_infos:
            screenshot_filepath = _make_screenshot(
                video_filepath=video_filepath,
                screenshot_filename_base=screenshot_filename_base,
                timestamp=timestamp,
                output_dir=output_dir,
                overwrite=overwrite,
            )
            output_queue.put((utils.daemon.MsgType.info, {
                'screenshot_filepath': screenshot_filepath,
                'video_filepath': video_filepath,
                'source': source,
            }))
    except errors.ScreenshotError as e:
        output_queue.put((utils.daemon.MsgType.error, e))


_ALL_SOURCES_SELECTED = '_ALL_SOURCES_SELECTED'

def _get_screenshot_infos(input_queue, *, custom_timestamps, count, exclude_files, from_all_videos):
    screenshot_infos = []
    while True:
        source = utils.daemon.read_input_queue_key(input_queue, 'source')
        if source == '_ALL_SOURCES_SELECTED':
            break
        else:
            # Return sequence of tuples where each contains all the information to create one screenshot.
            video_infos = _get_video_infos(source, exclude_files, from_all_videos)

            # We want to create `len(validated_timestamps)` screenshots for each item in `video_infos`.
            for source, video_filepath, screenshot_filename_base in video_infos:
                validated_timestamps = _validate_timestamps(
                    video_filepath=video_filepath,
                    timestamps=custom_timestamps,
                    count=count,
                )
                screenshot_infos.extend(
                    (source, video_filepath, screenshot_filename_base, ts)
                    for ts in validated_timestamps
                )

    return tuple(screenshot_infos)


def _get_video_infos(source, exclude_files, from_all_videos):
    if isinstance(source, utils.disc.Playlist):
        return _get_video_infos_from_playlist(source)
    else:
        return _get_video_infos_from_file_or_directory(source, exclude_files, from_all_videos)


def _get_video_infos_from_playlist(playlist):
    # Include parent directory name.
    screenshot_filename_base = [playlist.discname]

    # Only include the playlist filename for BDMVs. For DVDs, it is implied by the file name of the
    # playlist item.
    if not playlist.filename.startswith('VTS_'):
        screenshot_filename_base.append(utils.fs.strip_extension(playlist.filename))

    # Include file name of playlist item (e.g. 00003.m2ts or VTS_12_1.VOB) without file extension.
    screenshot_filename_base.append(utils.fs.strip_extension(utils.fs.basename(playlist.largest_item)))

    return (
        # (<source>, <video_filepath>, <screenshot_filename_base>)
        (playlist, playlist.largest_item, '.'.join(screenshot_filename_base)),
    )


def _get_video_infos_from_file_or_directory(content_path, exclude_files, from_all_videos):
    video_filepaths = utils.fs.find_main_videos(content_path, exclude_files)
    video_infos = []
    for video_filepath in video_filepaths:
        screenshot_filename_base = utils.fs.strip_extension(utils.fs.basename(video_filepath))
        video_infos.append((content_path, video_filepath, screenshot_filename_base))
        if not from_all_videos:
            break
    return tuple(video_infos)


def _make_screenshot(*, video_filepath, screenshot_filename_base, timestamp, output_dir, overwrite):
    screenshot_filepath = os.path.join(
        output_dir,
        screenshot_filename_base + f'.{timestamp}.png',
    )
    if not overwrite and os.path.exists(screenshot_filepath):
        return screenshot_filepath
    else:
        return utils.image.screenshot(
            video_file=video_filepath,
            screenshot_file=screenshot_filepath,
            timestamp=timestamp,
        )


def _validate_timestamps(*, video_filepath, timestamps, count):
    # Validate, normalize, deduplicate and sort timestamps

    duration = utils.mediainfo.get_duration(video_filepath)
    if duration < 1:
        raise errors.ContentError(f'Video duration is too short: {duration}s')

    # Convert timestamp int/float/str to Timestamp and limit its value.
    validated_timestamps = []
    min_ts = utils.types.Timestamp(0)
    max_ts = utils.types.Timestamp(duration)
    for ts in timestamps:
        try:
            ts = max(min_ts, min(max_ts, utils.types.Timestamp(ts)))
        except ValueError as e:
            raise errors.ContentError(e) from e
        else:
            validated_timestamps.append(ts)
    # Deduplicated validated_timestamps.
    validated_timestamps = sorted(set(validated_timestamps))

    if not timestamps and not count:
        count = DEFAULT_NUMBER_OF_SCREENSHOTS

    # Add more timestamps if the user didn't specify less than `count`.
    if count > 0 and len(validated_timestamps) < count:
        # Get position as fraction of video duration for each timestamp:
        # 0.0 = Timestamp(0)
        # 1.0 = Timestamp(duration)
        positions = [ts / duration for ts in sorted(validated_timestamps)]

        # Include start and end of video. They are required for the algorithm below.
        if 0.0 not in positions:
            positions.insert(0, 0.0)
        if 1.0 not in positions:
            positions.append(1.0)

        # Sort positions so they can be paired to find the largest gap.
        positions.sort()

        # Add new positions between the two positions with the largest gap until we have the desired
        # number of screenshots.
        while len(validated_timestamps) < count:
            pairs = zip(positions, positions[1:])
            _max_distance, pos1, pos2 = max((b - a, a, b) for a, b in pairs)
            new_position = ((pos2 - pos1) / 2) + pos1
            validated_timestamps.append(utils.types.Timestamp(int(duration * new_position)))
            positions.append(new_position)
            positions.sort()

    # Return deduplicated, sorted, immutable validated_timestamps.
    return tuple(sorted(set(validated_timestamps)))


def _optimize_process(
        output_queue, input_queue,
        *,
        level, overwrite, ignore_dependency_error, cache_directory,
):
    # Keep reading queued screenshots forever. read_input_queue_until_empty() raises
    # DaemonProcessTerminated if there's a `MsgType.terminate` queued up.
    msgs = []
    while True:
        new_msgs = utils.daemon.read_input_queue_until_empty(input_queue)
        msgs.extend(new_msgs)
        if msgs:
            _typ, info = msgs.pop(0)
            _optimize_screenshot(
                output_queue=output_queue,
                screenshot_filepath=info['screenshot_filepath'],
                video_filepath=info['video_filepath'],
                source=info['source'],
                level=level,
                overwrite=overwrite,
                ignore_dependency_error=ignore_dependency_error,
                cache_directory=cache_directory,
            )


def _optimize_screenshot(
        output_queue,
        *,
        screenshot_filepath, video_filepath, source,
        level, overwrite, ignore_dependency_error, cache_directory,
):
    output_file = utils.fs.ensure_path_in_cache(
        os.path.join(
            utils.fs.dirname(screenshot_filepath),
            (
                utils.fs.basename(utils.fs.strip_extension(screenshot_filepath))
                + '.'
                + f'optimized={level}'
                + '.'
                + utils.fs.file_extension(screenshot_filepath)
            )
        ),
        cache_directory,
    )

    if not overwrite and os.path.exists(output_file):
        output_queue.put((utils.daemon.MsgType.info, {
            'optimized_screenshot_filepath': output_file,
            'video_filepath': video_filepath,
            'source': source,
        }))

    else:
        try:
            optimized_screenshot_filepath = utils.image.optimize(
                screenshot_filepath,
                level=level,
                output_file=output_file,
            )

        except errors.ImageOptimizeError as e:
            output_queue.put((utils.daemon.MsgType.error, e))

        except errors.DependencyError as e:
            if ignore_dependency_error:
                # Act like we optimized `screenshot_filepath`
                output_queue.put((utils.daemon.MsgType.info, {
                    'optimized_screenshot_filepath': screenshot_filepath,
                    'video_filepath': video_filepath,
                    'source': source,
                }))
            else:
                raise e

        else:
            output_queue.put((utils.daemon.MsgType.info, {
                'optimized_screenshot_filepath': optimized_screenshot_filepath,
                'video_filepath': video_filepath,
                'source': source,
            }))



"""
Get information from "BDMV" directory trees
"""

import collections
import os
import re

from ... import utils

natsort = utils.LazyModule(module='natsort', namespace=globals())


def is_dvd(content_path, *, multidisc=False):
    """
    Whether `content_path` contains a "VIDEO_TS" subdirectory

    Also look for "VIDEO_TS.IFO" exists directly in `content_path`.

    If `multidisc` is truthy, also look for a "VIDEO_TS" directory in any subdirectory, but not
    recursively.
    """
    if os.path.isdir(content_path):
        if os.path.isdir(os.path.join(content_path, 'VIDEO_TS')):
            return True

        if os.path.exists(os.path.join(content_path, 'VIDEO_TS.IFO')):
            return True

        if multidisc:
            for subdir in utils.fs.listdir(content_path):
                if os.path.isdir(os.path.join(content_path, subdir, 'VIDEO_TS')):
                    return True

    return False


def get_disc_paths(content_path):
    """
    Return sequence of subdirectories that contain a "VIDEO_TS" directory

    If `content_path` contains a "VIDEO_TS.IFO" file, return `content_path`.

    Return empty sequence by default.
    """
    if os.path.isdir(content_path):
        discpaths = []

        # Most likely single-disc release.
        if os.path.isdir(os.path.join(content_path, 'VIDEO_TS')):
            discpaths.append(content_path)

        # The .VOB files may also be in `content_path` with no "VIDEO_TS" directory.
        if os.path.exists(os.path.join(content_path, 'VIDEO_TS.IFO')):
            discpaths.append(content_path)

        # Find more discs in subdirectories.
        for name in utils.fs.listdir(content_path):
            subpath = os.path.join(content_path, name)
            if name != 'VIDEO_TS' and is_dvd(subpath):
                discpaths.append(subpath)

        return tuple(natsort.natsorted(discpaths))

    return ()


def get_playlists(discpath):
    """
    Return sequence of :class:`~.Playlist` instances from VIDEO_TS subdirectory

    Return empty sequence if no playlists are found.

    Playlists with a runtime of less than 3 minutes are ignored.

    Each playlist represents one Video Title Set, e.g. ``VTS_01_*``, ``VTS_02_*``, ``VTS_03_*``,
    etc. Each playlist's :attr:`.Playlist.filepath` is the .IFO file and the :attr:`.Playlist.items`
    are .VOB files.

    :param discpath: Path to directory that contains a "VIDEO_TS" subdirectory
    """
    video_ts_path = _get_video_ts_directory(discpath)
    if not video_ts_path:
        return ()

    playlists = collections.defaultdict(dict)
    vob_sets = collections.defaultdict(list)
    for filename in utils.fs.listdir(video_ts_path):
        match = re.search(r'^(VTS_\d+)_\d+\.[A-Z]+$', filename)
        if match:
            set_name = match.group(1)
            filepath = os.path.join(video_ts_path, filename)
            file_extension = utils.fs.file_extension(filename).lower()

            if file_extension == 'ifo':
                playlists[set_name]['filepath'] = filepath
                playlists[set_name]['discpath'] = discpath
                playlists[set_name]['duration'] = utils.mediainfo.get_duration_from_mediainfo(filepath)

            elif file_extension == 'vob':
                vob_sets[set_name].append(filepath)

    # Move Video Title Sets to each playlist's `items`.
    for set_name, items in vob_sets.items():
        playlists[set_name]['items'] = tuple(sorted(items))

    # Remove playlist with less than 3 minutes runtime.
    for set_name, playlist in tuple(playlists.items()):
        if playlist['duration'] < 180:
            del playlists[set_name]

    playlists = tuple(
        natsort.natsorted(
            (
                utils.disc.Playlist(**playlist)
                for playlist in playlists.values()
            ),
            key=lambda playlist: playlist.filepath,
        )
    )
    utils.disc.playlist.mark_main_playlists(playlists)
    return playlists


def _get_video_ts_directory(discpath):
    """
    Return the path of the directory that contains the DVD files

    This may be ``<discpath>/VIDEO_TS`` or ``<discpath>``.

    :raise errors.ContentError: if `discpath` does not contain DVD files
    """
    if os.path.isdir(os.path.join(discpath, 'VIDEO_TS')):
        return os.path.join(discpath, 'VIDEO_TS')
    elif os.path.exists(os.path.join(discpath, 'VIDEO_TS.IFO')):
        return discpath



"""
Concrete :class:`~.TrackerBase` subclass for UHD
"""

import asyncio
import re
import urllib

from ... import __project_name__, errors, utils
from ..base import TrackerBase
from . import rules
from .config import UhdTrackerConfig
from .jobs import UhdTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class UhdTracker(TrackerBase):
    name = 'uhd'
    label = 'UHD'
    torrent_source_field = '[UHDBits]'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. Login Credentials\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.username USERNAME\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.password PASSWORD\n'
        '\n'
        '{howto.next_section}. Announce URL (Optional)\n'
        '\n'
        '   The announce URL is fetched from the website on demand, but you can\n'
        '   also configure it explicitly.\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.announce_url ANNOUNCE_URL\n'
        '\n'
        '{howto.screenshots}\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = UhdTrackerConfig
    TrackerJobs = UhdTrackerJobs
    rules = rules

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _login_url(self):
        return urllib.parse.urljoin(self._base_url, '/login.php')

    @property
    def _logout_url(self):
        return urllib.parse.urljoin(self._base_url, '/logout.php')

    @property
    def _ajax_url(self):
        return urllib.parse.urljoin(self._base_url, '/ajax.php')

    @property
    def _upload_url(self):
        return urllib.parse.urljoin(self._base_url, '/upload.php')

    @property
    def _torrents_url(self):
        return urllib.parse.urljoin(self._base_url, '/torrents.php')

    async def _request(self, method, *args, error_prefix='', **kwargs):
        try:
            # `method` is "GET" or "POST"
            return await getattr(utils.http, method.lower())(
                *args,
                user_agent=True,
                cache=False,
                **kwargs,
            )
        except errors.RequestError as e:
            if error_prefix:
                raise errors.RequestError(f'{error_prefix}: {e}') from e
            else:
                raise e

    def _failed_to_find_error(self, doc, msg_prefix):
        filepath = f'{msg_prefix}.{self.name}.html'
        utils.html.dump(doc, filepath)
        raise RuntimeError(f'{msg_prefix}: No error message found (dumped HTML response to {filepath})')

    async def _login(self, *, tfa_otp=None):
        if not self.options.get('username'):
            raise errors.RequestError('Login failed: No username configured')
        elif not self.options.get('password'):
            raise errors.RequestError('Login failed: No password configured')

        post_data = {
            'username': self.options['username'],
            'password': self.options['password'],
            'two_step': '',  # 2FA
            'login': 'Log in',
        }

        await self._request(
            method='POST',
            url=self._login_url,
            data=post_data,
            error_prefix='Login failed',
        )

    async def confirm_logged_in(self):
        response = await self._request('GET', self._base_url)
        doc = utils.html.parse(response)
        auth_regex = re.compile(r'logout\.php\?.*\bauth=([0-9a-zA-Z]+)')
        logout_link_tag = doc.find('a', href=auth_regex)
        if logout_link_tag:
            logout_link_href = logout_link_tag['href']
            match = auth_regex.search(logout_link_href)
            self._auth = match.group(1)
        else:
            msg_prefix = 'Login failed'
            form_tag = doc.find('form', action='login.php')
            if form_tag:
                form_tag.table.extract()
                msg = utils.html.as_text(form_tag)
                raise errors.RequestError(f'{msg_prefix}: {msg}')
            else:
                self._failed_to_find_error(doc, msg_prefix)

    async def _logout(self):
        try:
            await self._request(
                method='GET',
                url=self._logout_url,
                params={'auth': self._auth},
                error_prefix='Logout failed',
            )
        finally:
            delattr(self, '_auth')

    async def get_announce_url(self):
        if self.options.get('announce_url'):
            return self.options['announce_url']
        elif not self.is_logged_in:
            raise RuntimeError('Cannot get announce URL from website if not logged in')
        else:
            response = await self._request(
                method='GET',
                url=self._upload_url,
            )
            doc = utils.html.parse(response)
            announce_url_tag = doc.find('input', value=re.compile(r'^https?://.*/announce\b'))
            if announce_url_tag:
                return announce_url_tag['value']
            else:
                cmd = f'{__project_name__} set trackers.{self.name}.announce_url YOUR_URL'
                raise errors.RequestError(f'Failed to find announce URL - set it manually: {cmd}')

    @utils.blocking_memoize
    async def get_uhd_info(self, imdb_id):
        """
        Get IMDb information from tracker

        :param imdb_id: IMDb ID

        The return value is a dictionary returned by ajax.php. It may contain
        arbitrary keys or be empty.
        """
        assert imdb_id, 'IMDb ID is not available yet'

        try:
            # If ajax.php doesn't have the info cached, it responds with
            # {"error": "1"}, in which case, we wait a decreasing number of
            # seconds before trying again. This is more or less what the website
            # JS is doing.

            # Example responses:
            # {"error":1,"message":" Submited to fetcher..."}
            # {"error":1,"message":" Fetching..."}
            # {'status': 'failure', 'error': 'rate limit exceeded'}

            # Wait 6 seconds after the first try, then 3 seconds for up to 4
            # minutes combined.
            for sleep in (6,) + (3,) * 20 * 4:
                uhd_info = await self._get_uhd_info(imdb_id)
                if str(uhd_info.get('error', '1')) == '0':
                    _log.debug('UHD info: %r', uhd_info)
                    return uhd_info
                if str(uhd_info.get('error')) not in ('1', '0'):
                    _log.debug('Failed to get UHD info for %s: %r', imdb_id, uhd_info['error'])
                    break
                else:
                    _log.debug('Still waiting for UHD info: %r', uhd_info)
                    await asyncio.sleep(sleep)

        except errors.RequestError as e:
            _log.debug('Failed to get UHD info for %s: %r', imdb_id, e)
            pass

        return {}

    async def _get_uhd_info(self, imdb_id):
        params = {
            'action': 'imdb_fetch',
            'imdbid': imdb_id,
        }
        _log.debug(
            'UHD info URL for %s: %s?%s',
            imdb_id,
            self._ajax_url,
            '&'.join(
                f'{key}={urllib.parse.quote_plus(value)}'
                for key, value in params.items()
            ),
        )
        response = await self._request(
            method='GET',
            url=self._ajax_url,
            params=params,
        )
        return response.json()

    async def upload(self, tracker_jobs):
        assert self.is_logged_in

        post_data = tracker_jobs.post_data
        post_data['auth'] = self._auth

        _log.debug('POSTing data:')
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        post_files = tracker_jobs.post_files
        _log.debug('POSTing files: %r', post_files)

        response = await self._request(
            method='POST',
            url=self._upload_url,
            data=post_data,
            files=post_files,
            follow_redirects=False,
        )
        return self._handle_upload_response(response)

    def _handle_upload_response(self, response):
        _log.debug('############################### HANDLING UPLOAD RESPONSE ##########################')
        # "Location" header should contain the uploaded torrent's URL
        _log.debug('Upload response headers: %r', response.headers)
        location = response.headers.get('Location')
        _log.debug('Upload response location: %r', location)

        if location:
            torrent_page_url = urllib.parse.urljoin(self.options['base_url'], location)
            # Redirect URL should start with "https://.../torrents.php"
            if torrent_page_url.startswith(self._torrents_url):
                return torrent_page_url

        # Find error message in HTML
        msg_prefix = 'Upload failed'
        doc = utils.html.parse(response)
        error_tags = doc.select('#scontent .thin > p + p')
        if error_tags:
            msg = utils.html.as_text(error_tags[0])
            raise errors.RequestError(f'{msg_prefix}: {msg}')

        # Failed to find error message
        self._failed_to_find_error(doc, msg_prefix)



from ... import errors
from ..base import rules


class PtpBannedGroup(rules.BannedGroup):
    banned_groups = {
        'aXXo',
        'BMDRu',
        'BRrip',
        'CM8',
        'CrEwSaDe',
        'CTFOH',
        'd3g',
        'DNL',
        'FaNGDiNG0',
        'HD2DVD',
        'HDT',
        'HDTime',
        'ION10',
        'iPlanet',
        'KiNGDOM',
        'LAMA',
        'mHD',
        'mSD',
        'NhaNc3',
        'nHD',
        'nikt0',
        'nSD',
        'OFT',
        'PRODJi',
        'SANTi',
        'SasukeducK',
        'SPiRiT',
        'STUTTERSHIT',
        'ViSION',
        'VXT',
        'WAF',
        'WORLD',
        'x0r',
        'YIFY',
    }

    async def _check_custom(self):
        # No EVO encodes. WEB-DLs are fine.
        if (
                self.is_group('EVO')
                and 'WEB' not in self.release_name.source
        ):
            raise errors.BannedGroup('EVO', additional_info='No encodes, only WEB-DL')



"""
Connect the user interface to the engine room
"""

from .base import JobBase  # isort:skip
from .jobrunner import JobRunner  # isort:skip
from . import (
    bdinfo,
    custom,
    dialog,
    imghost,
    mediainfo,
    playlists,
    poster,
    rules,
    scene,
    screenshots,
    set,
    submit,
    torrent,
    webdb,
)



"""
Query online databases like IMDb
"""

import asyncio
import collections

from .. import errors
from ..utils import webdbs
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class WebDbSearchJob(JobBase):
    """
    Ask user to select a specific search result from an internet database

    This job adds the following signals to :attr:`~.JobBase.signal`:

        ``search_results``
            Emitted after new search results are available. Registered callbacks
            get a sequence of :class:`~.utils.webdbs.common.SearchResult`
            instances as a positional argument.

        ``searching_status``
            Emitted when a new search is started and when new results are
            available. Registered callbacks get `True` for "search started" or
            `False` for "search ended" as a positional argument.

        ``info_updating``
            Emitted when an attempt to fetch additional information is made.
            Registered callbacks get the name of an attribute of a
            :class:`~.utils.webdbs.common.SearchResult` object that returns a
            coroutine function as a positional argument.

        ``info_updated``
            Emitted when additional information is available. Registered
            callbacks are called for each piece of information and get ``key``
            and ``value`` as positional arguments. ``key`` is an attribute of a
            :class:`~.utils.webdbs.common.SearchResult` object that returns a
            coroutine function. ``value`` is the return value of that coroutine
            function.

        ``query_updated``
            Emitted after the :attr:`query` changed. Registered callbacks get
            the new query as a positional argument.

        ``selected``
            Emitted when :meth:`result_selected` is called. Registered callbacks
            get a :class:`dict` with the keys ``id``, ``title``, ``type``,
            ``url`` and ``year``.
    """

    @property
    def name(self):
        return f'{self._db.name}-id'

    @property
    def label(self):
        return f'{self._db.label} ID'

    @property
    def cache_id(self):
        """Database :attr:`~.WebDbApiBase.name` and :attr:`query`"""
        return self.query

    @property
    def db(self):
        """:class:`~.webdbs.base.WebDbApiBase` instance"""
        return self._db

    @property
    def query(self):
        """:class:`~.webdbs.common.Query` instance"""
        return self._query

    @property
    def no_id_ok(self):
        """Whether this job will finish gracefully if no ID is selected"""
        return self._no_id_ok

    @no_id_ok.setter
    def no_id_ok(self, value):
        self._no_id_ok = bool(value)

    @property
    def show_poster(self):
        """Whether to display a poster of the focused search result"""
        return self._show_poster

    @show_poster.setter
    def show_poster(self, value):
        self._show_poster = bool(value)

    def _get_show_poster(self):
        return self._show_poster

    @property
    def is_searching(self):
        """Whether a search request is currently being made"""
        return self._is_searching

    def initialize(self, *, db, query, autodetect=None, no_id_ok=False, show_poster=True):
        """
        Set internal state

        :param WebDbApiBase db: Return value of :func:`.utils.webdbs.webdb`
        :param query: Path of the release (doesn't have to exist) or
            :class:`~.webdbs.common.Query` instance
        :param autodetect: Coroutine function that takes no arguments and
            returns an ID or `None`

            If `autodetect` returns a valid ID it is automatically selected and
            the job is finished with no user interaction. Otherwise, the
            returned ID is inserted into the usual prompt and it's up to the
            user to fix it.
        :param no_id_ok: Whether this job may finish without error if no ID is
            selected
        :param show_poster: Whether to display a poster of the focused search
            result
        """
        assert isinstance(db, webdbs.WebDbApiBase), f'Not a WebDbApiBase: {db!r}'
        self._db = db
        if not isinstance(query, webdbs.Query):
            query = webdbs.Query.from_any(query)
        self._query = self._db.sanitize_query(query)
        self._autodetect = autodetect

        self.signal.add('search_results')
        self.signal.add('searching_status')
        self.signal.add('info_updating')
        self.signal.add('info_updated')
        self.signal.add('query_updated')
        self.signal.add('selected', record=True)

        # Call search() when `self.query` is changed externally, e.g. by another
        # job. This happens if we have multiple WebDbSearchJobs with different
        # WebDbs and we want to update the query of the second job with accurate
        # information from the first job. This means the user only has to fix a
        # potentially badly autodetected query once and not for each
        # WebDbSearchJob.
        self._query.signal.register('changed', self.search)

        # Hold some information about the selected search result.
        self._selected = {}
        self.signal.register('selected', self._selected.update)

        self.no_id_ok = no_id_ok
        self.show_poster = show_poster
        self._is_searching = False
        self._search_task = None
        self._info_callbacks_task = None
        self._info_callbacks = _InfoCallbacks(
            callbacks={
                'id': self._make_update_info_func('id'),
                'summary': self._make_update_info_func('summary'),
                'title_original': self._make_update_info_func('title_original'),
                'title_english': self._make_update_info_func('title_english'),
                'genres': self._make_update_info_func('genres'),
                'directors': self._make_update_info_func('directors'),
                'cast': self._make_update_info_func('cast'),
                'countries': self._make_update_info_func('countries'),
                'poster': self._make_update_info_func('poster', condition=self._get_show_poster),
            },
            error_callback=self.warn,
        )

    def _make_update_info_func(self, key, condition=lambda: True):
        def update_info_func(value):
            if condition():
                self._update_info_value(key, value)
        return update_info_func

    def _update_info_value(self, attr, value):
        if value is Ellipsis:
            # Indicate that we are updating `attr` by not providing a value.
            self.signal.emit('info_updating', attr)
        else:
            self.signal.emit('info_updated', attr, value)

    async def run(self):
        if self._autodetect is not None:
            id = await self._autodetect()
            _log.debug(f'Autodetected {self.db.name} ID from %r: %r', self._autodetect, id)
            if id:
                id = self._db.get_id_from_text(id)
                if id:
                    self.query.id = id
                    self.query.feeling_lucky = True

        # Trigger initial search. We must always search, even if an ID was
        # autodetected, to verify it's a valid ID.
        self.search(self.query)

        # Block until finalize() is called.
        await self.finalization()

    async def _search(self, query):
        self._set_state(is_searching=True, results=())

        try:
            # Get new search results
            results = await self._db.search(query)
            self._set_state(is_searching=False, results=results)

            if results:
                if query.feeling_lucky and len(results) == 1:
                    # Select single search result.
                    self.result_selected(results[0])
                else:
                    # Display details about the first search result.
                    self._run_info_callbacks(results[0])
            else:
                # Unset all details from previously selected search result.
                self._run_info_callbacks(None)

        except errors.RequestError as e:
            self.warn(e)

        finally:
            self._set_state(is_searching=False)
            # We no longer feel lucky. If the query was lucky, there is no next
            # query. If the query was unlucky, do not autoselect any new query.
            query.feeling_lucky = False

    def _set_state(self, *, is_searching=None, results=None):
        if is_searching is not None:
            self._is_searching = bool(is_searching)
            self.signal.emit('searching_status', self._is_searching)
        if results is not None:
            self.signal.emit('search_results', results)

    def _run_info_callbacks(self, result):
        if self._info_callbacks_task:
            self._info_callbacks_task.cancel()
        self._info_callbacks_task = self.add_task(
            self._info_callbacks(result),
            # Auto-remove reference to task when it is done.
            callback=self._unset_info_callbacks_task,
        )

    def _unset_info_callbacks_task(self, task_):
        self._info_callbacks_task = None

    def _cancel_tasks(self):
        for task_attr in ('_search_task', '_info_callbacks_task'):
            task = getattr(self, task_attr, None)
            if task:
                task.cancel()
                setattr(self, task_attr, None)

    def search(self, query):
        """
        Make new search request

        Any currently ongoing search is cancelled.

        :param str query: Query that is first passed through
            :meth:`~.webdbs.Query.from_any` and then
            :meth:`~.webdbs.WebDbApiBase.sanitize_query`
        """
        # Stop any ongoing requests. They are outdated.
        self._cancel_tasks()

        # Remove any warnings from previous requests (e.g. network issues).
        self.clear_warnings()

        # Update search query silently (without emitting any signals).
        new_query = self._db.sanitize_query(webdbs.Query.from_any(query))
        self.query.update(new_query, silent=True)
        self.signal.emit('query_updated', self.query)

        # This method is called to fix the initial query. If this happens before
        # this job is started, we cannot add any tasks yet. But that's ok
        # because we already updated our internal query, which will be used for
        # the initial search if this job is started.
        if self.is_started:
            # Initiate new search in a separate task so we can cancel it.
            self._search_task = self.add_task(
                self._search(new_query),
                # Auto-remove reference to task when it is done.
                callback=self._unset_search_task,
            )

    def _unset_search_task(self, task_):
        self._search_task = None

    def result_focused(self, result):
        """
        Must be called by the UI when the user focuses a different search result

        :param SearchResult result: Focused search result or `None` if there are
            no results
        """
        self._run_info_callbacks(result)

    def result_selected(self, result):
        """
        Must be called by the UI when the user accepts a search result

        :param SearchResult result: Focused search result or `None` if there are
            no results and the user accepted that (see :attr:`no_id_ok`)
        """
        if not self.is_searching:
            if result is not None:
                self.signal.emit('selected', {
                    'id': result.id,
                    'title': result.title,
                    'type': result.type,
                    'url': result.url,
                    'year': result.year,
                })
                self.add_output(result.id)

            if result is not None or self.no_id_ok:
                self._cancel_tasks()
                self.finalize()

    @property
    def selected(self):
        """
        Some more information about the selected search result

        This is a :class:`dict` that is empty before :meth:`result_selected`
        is called. After :meth:`result_selected` is called, it contains the
        following keys:

            - ``id``
            - ``title``
            - ``type``
            - ``url``
            - ``year``
        """
        return self._selected.copy()

    @property
    def exit_code(self):
        """`0` if job was successful, `> 0` otherwise, None while job is not finished"""
        exit_code = super().exit_code
        if exit_code is not None:
            if exit_code != 0 and self.no_id_ok:
                return 0
            else:
                return exit_code


class _InfoCallbacks:
    def __init__(self, callbacks, error_callback):
        # `callbacks` maps names of SearchResult attributes to callbacks that
        # get the value of each attribute.
        #
        # For example, {"title": handle_title} means: Get the "title" attribute
        # from a search result and pass it to handle_title().
        #
        # Values of SearchResult attributes may also be coroutine functions that
        # return the value. In that case, call
        # handle_title(await search_result.title()).
        self._callbacks = callbacks
        self._error_callback = error_callback

    async def __call__(self, result):
        if not result:
            for callback in self._callbacks.values():
                callback('')
        else:
            self._update_text_values(result)
            await self._update_awaitable_values(result)

    def _update_text_values(self, result):
        # Update plain, non-callable values first before awaiting coroutines.
        for attr, callback in self._callbacks.items():
            value = getattr(result, attr)
            if not callable(value):
                callback(self._convert_value(value))

    async def _update_awaitable_values(self, result):
        # Update values that are coroutine functions that return the value.
        tasks = []
        for attr, callback in self._callbacks.items():
            value = getattr(result, attr)
            if callable(value):
                tasks.append(self._call_callback(
                    value_getter=value,
                    callback=callback,
                    cache_key=(result.id, attr),
                ))
        # Get all values concurrently
        await asyncio.gather(*tasks)

    _cache = {}
    _delay_between_updates = 0.5

    async def _call_callback(self, value_getter, callback, cache_key):
        cached_value = self._cache.get(cache_key, None)
        if cached_value is not None:
            callback(cached_value)
        else:
            # Wait for a bit before starting to load info. If the user rapidly
            # focuses different search results, we don't want to start lots of
            # requests every few fractions of a second.
            callback(Ellipsis)
            await asyncio.sleep(self._delay_between_updates)

            # Get attribute value from from search result
            try:
                value = await value_getter()
            except errors.RequestError as e:
                callback('')
                self._error_callback(e)
            else:
                value_converted = self._convert_value(value)
                self._cache[cache_key] = value_converted
                callback(value_converted)

    @staticmethod
    def _convert_value(value):
        if (
                isinstance(value, collections.abc.Iterable)
                and not isinstance(value, (str, bytes, bytearray))
        ):
            return ', '.join(str(v) for v in value)
        else:
            return value



"""
CLI argument types

All types return normalized values and raise ValueError for invalid values.
"""

import collections
import enum
import os
import re

from .. import utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def Integer(value, min=None, max=None):
    """
    Return :class:`int` subclass with minimum and maximum value

    >>> i = Integer(5, min=0, max=10)
    >>> type(i)(100)
    >>> ValueError: Maximum is 10
    """
    # There's a Python bug that prevents us from overloading min() and max()
    # with variables in the "class ...:" namespace
    min_ = min
    max_ = max

    class Integer(int):
        min = min_
        """Minimum value"""

        max = max_
        """Maximum value"""

        def __new__(cls, value):
            try:
                i = int(float(value))
            except (ValueError, TypeError) as e:
                raise ValueError(f'Invalid integer value: {value!r}') from e

            if cls.min is not None and i < cls.min:
                raise ValueError(f'Minimum is {cls.min}')
            elif cls.max is not None and i > cls.max:
                raise ValueError(f'Maximum is {cls.max}')
            else:
                return super().__new__(cls, i)

        def __str__(self):
            return str(int(self))

        def __repr__(self):
            string = f'{type(self).__name__}({super().__repr__()}'
            if min is not None:
                string += f', min={min!r}'
            if max is not None:
                string += f', max={max!r}'
            string += ')'
            return string

    return Integer(value)


def Choice(value, options, *, empty_ok=False, case_sensitive=True):
    """
    Return :class:`str` subclass that can only have instances that are equal to
    an item of `options`

    :param value: Initial value
    :param options: Iterable of allowed instances
    :param bool empty_ok: Whether an emptry string is valid even if it is not in
        `options`
    :param bool case_sensitive: Whether case is considered

    :raise ValueError: if instantiation is attempted with a value that is not in
        `options`
    """
    if case_sensitive:
        options_str = tuple(sorted(str(o) for o in options))
    else:
        options_str = tuple(sorted(str(o).lower() for o in options))

    class Choice(str):
        options = options_str

        def __new__(cls, val):
            val_str = str(val)
            if not case_sensitive:
                val_str = val_str.lower()

            if val_str not in cls.options and (val_str or not empty_ok):
                raise ValueError(f'Not one of {", ".join(cls.options)}: {val}')
            else:
                return super().__new__(cls, val)

        def __str__(self):
            return super().__str__()

        def __repr__(self):
            return f'{type(self).__name__}({super().__repr__()}, options={self.options!r})'

    return Choice(value)


def ImageHost(allowed=None, disallowed=None):
    """
    Return new :class:`Choice` subclass that only accepts `allowed`
    image host names

    :param allowed: Sequence of allowed image host names or `None`
        to allow all supported image host names
    :param disallowed: Sequence of disallowed image host names or `None`
        to allow all supported image host names
    """
    if allowed:
        options = set(allowed)
    else:
        options = set(utils.imghosts.imghost_names())

    if disallowed:
        for item in disallowed:
            options.discard(item)

    ImageHost = type(Choice(
        sorted(options)[0],
        options=options,
        case_sensitive=False,
    ))
    ImageHost.__name__ = 'ImageHost'
    ImageHost.__qualname__ = ImageHost.__name__
    ImageHost.__doc__ = "Name of a supported image hosting service"
    return ImageHost


class Bool(str):
    """
    :class:`str` subclass with boolean value

    Truthy strings: ``true``, ``yes``, ``on``, ``1``
    Falsy strings: ``false``, ``no``, ``off``, ``0``
    """

    truthy = ('true', 'yes', '1', 'on', 'aye')
    """Valid `True` values (case-insensitive)"""

    falsy = ('false', 'no', '0', 'off', 'nay')
    """Valid `False` values (case-insensitive)"""

    _truthy = re.compile(r'^(?:' + '|'.join(truthy) + ')$', flags=re.IGNORECASE)
    _falsy = re.compile(r'^(?:' + '|'.join(falsy) + ')$', flags=re.IGNORECASE)

    def __new__(cls, value):
        self = super().__new__(cls, value)
        if cls._truthy.search(self):
            self._bool = True
        elif cls._falsy.search(self):
            self._bool = False
        else:
            raise ValueError(f'Invalid boolean value: {value!r}')
        return self

    def __bool__(self):
        return self._bool

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return other._bool == self._bool
        elif isinstance(other, bool):
            return other == self._bool
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, type(self)):
            return other._bool != self._bool
        elif isinstance(other, bool):
            return other != self._bool
        else:
            return NotImplemented

    def __repr__(self):
        return f'{type(self).__name__}({super().__str__()!r})'


class Bytes(int):
    """:class:`int` subclass with binary or decimal unit prefix"""

    _regex = re.compile(r'^(\d+(?:\.\d+|)) ?([a-zA-Z]{,3})$')
    _multipliers = {
        '': 1,
        'k': 1000,
        'M': 1000**2,
        'G': 1000**3,
        'T': 1000**4,
        'P': 1000**5,
        'Ki': 1024,
        'Mi': 1024**2,
        'Gi': 1024**3,
        'Ti': 1024**4,
        'Pi': 1024**5,
    }

    @classmethod
    def from_string(cls, string):
        """Parse `string` like ``4kB`` or ``1.024 KiB``"""
        match = cls._regex.search(string)
        if not match:
            raise ValueError(f'Invalid size: {string}')
        else:
            number = match.group(1)
            unit = match.group(2)
            if unit and unit[-1] == 'B':
                unit = unit[:-1]
            try:
                multiplier = cls._multipliers[unit]
            except KeyError as e:
                raise ValueError(f'Invalid unit: {unit}') from e
            else:
                return cls(int(float(number) * multiplier))

    def __new__(cls, value):
        if isinstance(value, str):
            return cls.from_string(value)
        else:
            return super().__new__(cls, value)

    def format(self, *, prefix='shortest', decimal_places=2, trailing_zeros=False):
        """
        Return human-readable string

        :param str prefix: Unit prefix, must be one of ``binary`` (1000 -> "1
            kB"), ``decimal`` (1024 -> "1 KiB") or ``shortest`` (automatically
            pick the string representation with the fewest decimal places)
        :param int decimal_places: How many decimal places to include
        :param bool trailing_zeros: Whether to remove zeros on the right of the
            decimal places
        """
        decimal_multipliers = (
            (prefix, multiplier)
            for prefix, multiplier in reversed(tuple(self._multipliers.items()))
            if len(prefix) == 1
        )
        binary_multipliers = (
            (prefix, multiplier)
            for prefix, multiplier in reversed(tuple(self._multipliers.items()))
            if len(prefix) == 2
        )

        string_format = f'{{number:.{decimal_places}f}}'

        def get_string(prefix, multipliers):
            for prefix, multiplier in multipliers:
                if self >= multiplier:
                    number = strip_trailing_zeros(string_format.format(number=self / multiplier))
                    return f'{number} {prefix}B'
            number = strip_trailing_zeros(string_format.format(number=int(self)))
            return f'{number} B'

        def strip_trailing_zeros(string):
            # Only strip zeros from decimal places (not from "100")
            if not trailing_zeros and '.' in string:
                return string.rstrip('0').rstrip('.') or '0'
            else:
                return string

        def number_of_decimal_places(string):
            number = str(''.join(c for c in str(string) if c in '1234567890.'))
            if '.' in number:
                return len(number.split('.', maxsplit=1)[1])
            else:
                return 0

        if prefix == 'binary':
            return get_string(prefix, binary_multipliers)
        elif prefix == 'decimal':
            return get_string(prefix, decimal_multipliers)
        elif prefix == 'shortest':
            decimal_string = get_string(prefix, decimal_multipliers)
            binary_string = get_string(prefix, binary_multipliers)
            sorted_strings = sorted((decimal_string, binary_string),
                                    key=number_of_decimal_places)
            return sorted_strings[0]
        else:
            raise ValueError(f'Invalid prefix: {prefix!r}')

    def __str__(self):
        return self.format()

    def __repr__(self):
        return f'{type(self).__name__}({int(self)!r})'


class Timestamp(float):
    """
    Subclass of :class:`float` that can parse and format timestamp/duration strings

    :param seconds: Number of seconds or hours, minutes and seconds as ":"-separated string
    :type seconds: int or float or "[[H+:]M+:]S+"

    :raise ValueError: if `seconds` is not a valid timestamp
    :raise TypeError: if `seconds` is not a :class:`int`, :class:`float` or :class:`str`
    """
    def __new__(cls, seconds):
        if isinstance(seconds, str):
            return cls.from_string(seconds)
        elif not isinstance(seconds, (int, float)):
            raise TypeError(f'Not a string or number: {seconds!r}')
        elif seconds < 0:
            raise ValueError(f'Timestamp must not be negative: {seconds!r}')
        else:
            return super().__new__(cls, round(seconds, 3))

    def __str__(self):
        text = ':'.join((
            f'{int(self / 3600)}',
            f'{int(self % 3600 / 60):02d}',
            f'{int(self % 3600 % 60):02d}',
        ))
        decimal_places = round(self - int(self), 3)
        if decimal_places > 0:
            text += f'.{str(decimal_places)[2:]}'
        return text

    def __repr__(self):
        return f'{type(self).__name__}({str(self)!r})'

    @classmethod
    def from_string(cls, string):
        """
        Parse string of the format "[[H+:]MM:]SS"

        :param str string: Hours, minutes and seconds as ":"-separated string or number of seconds

        :raise ValueError: if `string` has an invalid format
        """
        try:
            parts = tuple(float(part) for part in str(string).split(':'))
        except ValueError as e:
            raise ValueError(f'Invalid timestamp: {string!r}') from e

        for part in parts:
            if part < 0:
                raise ValueError(f'Timestamp must not be negative: {string}')

        if len(parts) == 3:
            hours, mins, secs = parts
        elif len(parts) == 2:
            hours = 0
            mins, secs = parts
        elif len(parts) == 1:
            hours = mins = 0
            secs = parts[0]
        else:
            raise ValueError(f'Invalid timestamp: {string}')

        return cls((hours * 3600) + (mins * 60) + secs)


class ReleaseType(enum.Enum):
    """
    Enum with the values ``movie``, ``season``, ``episode`` and
    ``unknown``

    ``series`` is an alias for ``season``.

    All values are truthy except for ``unknown``.
    """

    movie = 'movie'
    season = 'season'
    series = 'season'
    episode = 'episode'
    unknown = 'unknown'

    def __bool__(self):
        return self is not self.unknown

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'{type(self).__name__}.{self.value}'


class SceneCheckResult(enum.Enum):
    """
    Enum with the values ``true``, ``false``, ``renamed``, ``altered`` and
    ``unknown``

    All values are falsy except for ``true``.
    """

    true = 'true'
    false = 'false'
    renamed = 'renamed'
    altered = 'altered'
    unknown = 'unknown'

    def __bool__(self):
        return self is self.true

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f'{type(self).__name__}.{self.value}'


class RegEx:
    """
    Special class that behaves like :class:`re.Pattern` (return value of
    :func:`re.compile`) but more intuitively

    Its string representation is :attr:`re.Pattern.pattern` instead of
    ``re.compile('<pattern>')``.

    Instead of raising :class:`re.error`, it raises :class:`ValueError` on
    invalid regular expressions.

    :raise ValueError: instead of :class:`re.error`
    """

    def __init__(self, pattern):
        if isinstance(pattern, type(self)):
            self._regex = pattern._regex
        else:
            try:
                self._regex = re.compile(pattern)
            except re.error as e:
                orig_msg = str(e)
                msg = orig_msg[0].upper() + orig_msg[1:]
                raise ValueError(f'{pattern}: {msg}') from e

    def __getattr__(self, name):
        return getattr(self._regex, name)

    def __str__(self):
        return self._regex.pattern

    def __repr__(self):
        return f'{type(self).__name__}({self._regex.pattern!r})'

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._regex.pattern == other._regex.pattern
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._regex)

    # Prevent maximum recursion error in __getattr__().
    # See: https://docs.python.org/3/library/pickle.html#object.__reduce__
    def __reduce__(self):
        return (type(self), (self._regex.pattern,))

    # If we copy.deepcopy() instances of this class (e.g. `ListOf(RegEx)` that
    # is used for trackers.*.exclude settings, we end up with `re.compile(...)`
    # instances for some reason. Implementing __deepcopy__() fixes that.
    def __deepcopy__(self, memo):
        return type(self)(self._regex.pattern)


def ListOf(item_type, default=(), separator=None):
    """
    Immutable list of `item_type` objects

    :param item_type: Any callable that returns a validated object for one of the
        list items or raises :class:`ValueError` or :class:`TypeError`
    :param default: Any iterable of initial items

        Items in `default` are converted to `item_type` as well.

    :raise ValueError: if any invalid value is added to the list

    :return: new instance of a subclass of :class:`~.collections.abc.Sequence`
    """

    # Avoid NameError bug (https://github.com/python/cpython/issues/87546)
    item_type_ = item_type
    separator_ = separator

    class ListOf(collections.abc.Sequence):

        item_type = item_type_
        separator = separator_

        @classmethod
        def _convert(cls, item):
            if isinstance(item, str):
                # Split `item` into multiple items
                if separator and separator in item:
                    for subitem in item.split(separator):
                        yield from cls._convert(subitem)

                # Ignore empty `item` if `separator` is specified. This
                # conveniently converts "foo,,bar,baz" to ["foo", "bar", "baz"].
                elif item or not separator:
                    try:
                        yield cls.item_type(item)
                    except (ValueError, TypeError) as e:
                        raise ValueError(f'Invalid value: {item}') from e

            # `item` is multiple values
            elif isinstance(item, collections.abc.Iterable):
                for subitem in item:
                    yield from cls._convert(subitem)

            # `item` is single value
            else:
                try:
                    yield cls.item_type(item)
                except (ValueError, TypeError) as e:
                    raise ValueError(f'Invalid value: {item}') from e

        def __new__(cls, items=()):
            self = super().__new__(cls)
            self._sequence = tuple(self._convert(items))
            return self

        def __getitem__(self, key):
            return self._sequence[key]

        def __len__(self):
            return len(self._sequence)

        def __eq__(self, other):
            if hasattr(other, 'item_type') and hasattr(other, '_sequence'):
                return (
                    self.item_type is other.item_type
                    and self._sequence == other._sequence
                )
            elif isinstance(other, collections.abc.Sequence):
                return self._sequence == tuple(other)
            else:
                return NotImplemented

        def __hash__(self):
            return hash((
                self._sequence,
                self.item_type,
            ))

        def __str__(self):
            return ', '.join(
                str(item)
                for item in self._sequence
            )

        def __repr__(self):
            return f'{type(self).__name__}({self._sequence!r})'

    # Specify class name for easier debugging (e.g. "ListOfSomeClassName")
    ListOf.__name__ = (
        'ListOf'
        + item_type.__name__[0].upper()
        + item_type.__name__[1:]
    )
    ListOf.__qualname__ = ListOf.__name__
    ListOf.__doc__ = f"""Immutable list of {item_type_.__qualname__}"""

    # Instantiate custom list with provided items
    return ListOf(default)


class PathTranslation(collections.namedtuple('PathTranslation', ('local', 'remote'))):
    """
    :class:`collections.namedtuple` that takes two paths separated by ``->``
    and provides them as the attributes :attr:`local` and :attr:`remote`.
    """

    _regex = re.compile(r'^\s*(.+?)\s*->\s*(.+?)\s*$')

    def __new__(cls, string):
        match = cls._regex.search(string)
        if match:
            local, remote = match.groups()
            return super().__new__(cls, local.rstrip(os.sep), remote.rstrip(os.sep))
        else:
            raise ValueError(f'Invalid path translation: {string}')

    def __str__(self):
        return f'{self.local} -> {self.remote}'


class PathTranslations(type(ListOf(PathTranslation))):
    """
    :class:`~.types.ListOf` :class:`PathTranslation` instances
    """

    def translate(self, path):
        """
        Replace each occurrence of `local` with `remote` in `path` for each
        :class:`PathTranslation` and return the resulting path.
        """
        for translation in self:
            _log.debug('Translating path: %r.replace(%r, %r) = %r',
                       path, translation.local, translation.remote,
                       path.replace(translation.local, translation.remote))
            path = path.replace(translation.local, translation.remote)
        return path



"""
Concrete :class:`~.TrackerConfigBase` subclass for ANT
"""

import base64

from ... import utils
from ..base import TrackerConfigBase, exclude


class AntTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly9hbnRoZWxpb24ubWU=').decode('ascii'),
        'apikey': utils.configfiles.config_value(
            value='',
            description='Your personal upload API key you created in your profile.',
        ),
        'announce_url': utils.configfiles.config_value(
            value='',
            description='Your personal announce URL.',
        ),
        'exclude': (
            exclude.checksums,
            exclude.images,
            exclude.nfo,
            exclude.samples,
        ),
        'anonymous': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description='Whether your username is displayed on your uploads.',
        ),
    }

    argument_definitions = {
        'submit': {
            ('--anonymous', '--an'): {
                'help': 'Hide your username for this submission',
                'action': 'store_true',
                # This must be `None` so it doesn't override the "anonymous"
                # value from the config file. See CommandBase.get_options().
                'default': None,
            },
            ('--nfo',): {
                'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
            },
            ('--tmdb', '--tm'): {
                'help': 'TMDb ID or URL',
                'type': utils.argtypes.webdb_id('tmdb'),
            },
        },
    }



"""
Concrete :class:`~.TrackerBase` subclass for MTV
"""

import re
import urllib

from ... import __project_name__, errors, utils
from ..base import TrackerBase
from . import rules
from .config import MtvTrackerConfig
from .jobs import MtvTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MtvTracker(TrackerBase):
    name = 'mtv'
    label = 'MTV'
    torrent_source_field = 'MTV'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. Login Credentials\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.username USERNAME\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.password PASSWORD\n'
        '\n'
        '{howto.next_section}. Announce URL (Optional)\n'
        '\n'
        '   The announce URL is fetched from the website on demand, but you can\n'
        '   also configure it explicitly.\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.announce_url ANNOUNCE_URL\n'
        '\n'
        '{howto.screenshots}\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = MtvTrackerConfig
    TrackerJobs = MtvTrackerJobs
    rules = rules

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _login_url(self):
        return urllib.parse.urljoin(self._base_url, '/login')

    @property
    def _login_tfa_url(self):
        return urllib.parse.urljoin(self._base_url, '/twofactor/login')

    @property
    def _logout_url(self):
        return urllib.parse.urljoin(self._base_url, '/logout')

    @property
    def _upload_url(self):
        return urllib.parse.urljoin(self._base_url, '/upload.php')

    @property
    def _torrents_url(self):
        return urllib.parse.urljoin(self._base_url, '/torrents.php')

    async def _request(self, method, *args, **kwargs):
        return await getattr(utils.http, method.lower())(
            *args,
            user_agent=True,
            cookies=self.cookies_filepath,
            follow_redirects=False,
            **kwargs,
        )

    async def _login(self, *, tfa_otp=None):
        if not self.options.get('username'):
            raise errors.RequestError('Login failed: No username configured')
        elif not self.options.get('password'):
            raise errors.RequestError('Login failed: No password configured')

        if not tfa_otp:
            # First request sends username and password.
            request_kwargs = {
                'url': self._login_url,
                'data': {
                    'token': await self._get_token(self._login_url),
                    'username': self.options['username'],
                    'password': self.options['password'],
                    # Tell the server to remember our user session if the user wants to store it.
                    'keeploggedin': '1' if self.cookies_filepath else None,
                    # Lock session to our IP address unless we have a long-running user session
                    # during which our IP address may change.
                    'iplocked': None if self.cookies_filepath else '1',
                    'cinfo': '1280|720|24|0',
                    'submit': 'login',
                },
            }
        else:
            # Second request sends 2FA OTP.
            request_kwargs = {
                'url': self._login_tfa_url,
                'data': {
                    'token': await self._get_token(self._login_tfa_url),
                    'code': tfa_otp,
                    'submit': 'login'
                },
            }

        response = await self._request('POST', **request_kwargs)
        if response.headers.get('location', '').endswith('twofactor/login'):
            raise errors.TfaRequired('2FA OTP required')

    async def confirm_logged_in(self):
        response = await self._request('GET', self._base_url)
        doc = utils.html.parse(response)
        user_page_link_regex = re.compile(r'/user.php(?:\?|.*?&)id=(\d+)')
        user_page_link_tag = doc.find('a', href=user_page_link_regex)
        if not user_page_link_tag:
            raise errors.RequestError('Login failed')

    async def _logout(self):
        token = await self._get_token(self._base_url)
        await self._request(
            'POST',
            url=self._logout_url,
            data={'token': token},
        )

    async def get_announce_url(self):
        if self.options.get('announce_url'):
            return self.options['announce_url']
        elif not self.is_logged_in:
            raise RuntimeError('Cannot get announce URL from website if not logged in')
        else:
            response = await self._request('GET', self._upload_url)
            doc = utils.html.parse(response)
            announce_url_tag = doc.find('input', value=re.compile(r'^https?://.*/announce\b'))
            if announce_url_tag:
                return announce_url_tag['value']
            else:
                cmd = f'{__project_name__} set trackers.{self.name}.announce_url YOUR_URL'
                raise errors.RequestError(f'Failed to find announce URL - set it manually: {cmd}')

    @staticmethod
    def calculate_piece_size_min_max(_bytes):
        """Anything from 32 KiB to 8 MiB is fine, regardless of content size"""
        return (
            32 * 1024,        # 32 KiB
            8 * (1024 ** 2),  # 8 MiB
        )

    async def upload(self, tracker_jobs):
        _log.debug('Initial POST data:\n')
        for k, v in tracker_jobs.post_data_upload.items():
            _log.debug(' * %s = %s', k, v)

        autofill_post_data = await self._make_autofill_request(tracker_jobs)
        _log.debug('Autofill data: %r', autofill_post_data)

        torrent_page_url = await self._make_upload_request(tracker_jobs, autofill_post_data)
        _log.debug('Torrent page URL: %r', torrent_page_url)
        return torrent_page_url

    async def _make_autofill_request(self, tracker_jobs):
        # First request uploads the torrent file. We extract "tempfileid" and
        # "tempfilename" from the returned HTML form that must be returned in
        # the second request. We can also extract the autogenerated "taglist".
        post_data = await self._prepare_post_data(tracker_jobs.post_data_autofill)
        _log.debug('Autofill POST data: %r', post_data)
        post_files = {
            'file_input': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }

        response = await self._request(
            'POST',
            url=self._upload_url,
            cache=False,
            data=post_data,
            files=post_files,
        )

        doc = utils.html.parse(response)
        try:
            return {
                'tempfileid': self._get_form_value(doc, 'input', attrs={'name': 'tempfileid'}),
                'tempfilename': self._get_form_value(doc, 'input', attrs={'name': 'tempfilename'}),
                'taglist': self._get_form_value(doc, 'textarea', attrs={'name': 'taglist'}),
            }
        except ValueError as e:
            # If we can't find required values, look for error message
            _log.debug('Failed to extract values from autofill response: %s', e)
            self._raise_error(doc, msg_prefix='Upload failed', tracker_jobs=tracker_jobs)

    async def _make_upload_request(self, tracker_jobs, autofill_post_data):
        # Second request combines our metadata with server-generated data from
        # _make_autofill_request().
        post_data = await self._prepare_post_data(tracker_jobs.post_data_upload)
        post_data.update(autofill_post_data)

        _log.debug('Final POST data:\n')
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        response = await self._request(
            'POST',
            url=self._upload_url,
            cache=False,
            data=post_data,
        )

        # Get URL to uploaded torrent from HTTP 302 redirect location
        redirect_path = response.headers.get('location', '')
        _log.debug('HTTP 302 redirect location: %r', redirect_path)
        if 'torrents.php' in redirect_path:
            torrent_page_url = urllib.parse.urljoin(self._base_url, redirect_path)
            return torrent_page_url

        # Find error message in HTML
        doc = utils.html.parse(response)
        self._raise_error_dupes(doc)
        try:
            self._raise_error(doc, msg_prefix='Upload failed', tracker_jobs=tracker_jobs)
        except RuntimeError:
            # _raise_error() can't find an error message.
            # Dump response headers for debugging.
            # HTML was already dumped by _raise_error()
            headers_filepath = f'{utils.fs.basename(tracker_jobs.content_path)}.headers'
            with open(headers_filepath, 'w') as f:
                for k, v in response.headers.items():
                    f.write(f'{k}: {v}\n')
            raise

    async def _prepare_post_data(self, post_data):
        # Remove `None` values from post_data values and ensure all keys and
        # values are strings.
        post_data = {
            str(k): str(v)
            for k, v in post_data.items()
            if v is not None
        }
        post_data['auth'] = await self._get_auth()
        return post_data

    async def _get_token(self, url):
        # Return hidden <input name="token"> tag's value.
        response = await self._request('GET', url)
        doc = utils.html.parse(response)
        return self._get_form_value(doc, 'input', attrs={'name': 'token'})

    async def _get_auth(self):
        # "auth" seems to be a random string that never changes, and it must be
        # returned in upload requests.
        if not hasattr(self, '_auth'):
            assert self.is_logged_in
            response = await self._request('GET', self._upload_url)
            doc = utils.html.parse(response)

            # There is a limit for unchecked torrents. Of course, it is not presented like any other
            # error message and there is no straightforward way to find in the HTML soup.
            tags = doc.select('#upload #content > div > div > h2')
            if tags:
                text = utils.html.as_text(tags[0])
                if re.search(r'\d+ torrents awaiting staff approval', text):
                    raise errors.RequestError(text)

            self._auth = self._get_form_value(doc, 'input', attrs={'name': 'auth'})

        return self._auth

    def _get_form_value(self, doc, tag_name, **kwargs):
        try:
            tag = doc.find(tag_name, **kwargs)

            # <input value="...">
            value = tag.get('value')
            if value:
                return value

            # <textarea>value</textarea>
            value = utils.html.as_text(tag)
            if value:
                return value

            raise ValueError(f'Tag has no value: {tag}')

        except AttributeError:
            pass

        raise ValueError(f'Could not find tag: {tag_name}')

    def _raise_error(self, doc, msg_prefix, tracker_jobs=None):
        error_tag = doc.find('div', attrs={'class': 'error'})
        if error_tag:
            msg = utils.html.as_text(error_tag)
            raise errors.RequestError(f'{msg_prefix}: {msg}')

        # Example error: "The exact same torrent file already exists on the site!"
        alert_tag = doc.find('div', attrs={'class': 'alert'})
        if alert_tag:
            msg = utils.html.as_text(alert_tag)
            raise errors.RequestError(f'{msg_prefix}: {msg}')

        # TODO: Remove this if the "zero bytes response / GroupID cannot be null" bug is fixed.
        #       Also remember to remove the then unnecessary `tracker_jobs` argument.
        if not doc.string and tracker_jobs:
            _log.debug('Empty server response: %r', doc.prettify())
            bug_report = ' '.join((
                tracker_jobs.release_name.title_with_aka_and_year,
                f'https://imdb.com/title/{tracker_jobs.imdb_id}',
            ))
            forum_link = f'{self._base_url}/forum/thread/3338'
            raise errors.RequestError(
                f'{msg_prefix}: "GroupID cannot be null" bug encountered.\n'
                '\n'
                f'Please post the following information here: {forum_link}\n'
                '\n'
                f'    {bug_report}\n'
                '\n'
                f'You will get a reply from staff when the issue is fixed and you can try again.\n'
                '\n'
                'Here is an owl for your inconvenience:\n'
                '\n'
                '   ^ ^\n'
                '  (O,O)\n'
                '\\ (   )   _,~´\n'
                ' `~"-"~~~´\n'
            )

        filepath = f'{msg_prefix}.{self.name}.html'
        utils.html.dump(doc, filepath)
        raise RuntimeError(f'{msg_prefix}: No error message found (dumped HTML response to {filepath})')

    def _raise_error_dupes(self, doc):
        dupes_warning = doc.find('div', id="messagebar", string=re.compile(r'(?i:dupes?|duplicates?)'))
        if dupes_warning:
            _log.debug('Found dupes warning: %s', dupes_warning)
            table = doc.find('table', id='torrent_table')
            if table:
                dupe_files = tuple(
                    utils.html.as_text(cell)
                    for row in table.find_all('tr', attrs={'class': 'torrent'})
                    for cell in row.find('td', attrs={'class': 'torrent'})
                )
                if dupe_files:
                    raise errors.FoundDupeError(dupe_files)
